#!/usr/bin/env python
import time
import sys

file_template = """
/* generated on %(when)s */

enum _oursqlx_exception_type _oursqlx_exc_from_errno(int err) {
    switch (err) {
%(code)s
        default:
            if (err >= ER_ERROR_FIRST && err <= ER_ERROR_LAST)
                return _oursqlx_ProgrammingError;
            else if (err > CR_MIN_ERROR && err < CR_MAX_ERROR)
                return _oursqlx_InterfaceError;
    }
    return _oursqlx_UnknownError;
}
"""

case_template = """
#ifdef %(const)s
        case %(const)s:
#endif
"""
return_template = """
            return %(exc)s;
"""

data = dict(
    _oursqlx_DataError="""
        CR_DATA_TRUNCATED
        ER_DIVISION_BY_ZERO
        ER_DUPLICATED_VALUE_IN_TYPE
        ER_TRUNCATED_WRONG_VALUE
        ER_TRUNCATED_WRONG_VALUE_FOR_FIELD
        ER_WARN_DATA_OUT_OF_RANGE
        ER_WARN_NULL_TO_NOTNULL
        ER_WRONG_VALUE
    """,
    _oursqlx_IntegrityError="""
        ER_DROP_INDEX_FK
        ER_DUP_ENTRY
        ER_DUP_ENTRY_AUTOINCREMENT_CASE
        ER_DUP_UNIQUE
        ER_FOREIGN_DUPLICATE_KEY
        ER_NO_REFERENCED_ROW
        ER_NO_REFERENCED_ROW_2
        ER_ROW_IS_REFERENCED
        ER_ROW_IS_REFERENCED_2
    """,
    _oursqlx_InternalError="""
        ER_DUP_KEY
        ER_FILSORT_ABORT
        ER_KEY_NOT_FOUND
    """,
    _oursqlx_NotSupportedError="""
        CR_NOT_IMPLEMENTED
        ER_CHECK_NOT_IMPLEMENTED
        ER_DELAYED_NOT_SUPPORTED
        ER_DUMP_NOT_IMPLEMENTED
        ER_FOREIGN_KEY_ON_PARTITIONED
        ER_ILLEGAL_HA
        ER_ILLEGAL_HA_CREATE_OPTION
        ER_ILLEGAL_REFERENCE
        ER_NDB_CANT_SWITCH_BINLOG_FORMAT
        ER_NOT_SUPPORTED_AUTH_MODE
        ER_NOT_SUPPORTED_YET
        ER_NO_RAID_COMPILED
        ER_NULL_COLUMN_IN_INDEX
        ER_PARTITION_FUNCTION_FAILURE
        ER_SLAVE_IGNORED_SSL_PARAMS
        ER_SP_NO_AGGREGATE
        ER_SP_SUBSELECT_NYI
        ER_TABLESPACE_AUTO_EXTEND_ERROR
        ER_TABLE_CANT_HANDLE_AUTO_INCREMENT
        ER_TABLE_CANT_HANDLE_BLOB
        ER_TABLE_CANT_HANDLE_FT
        ER_TABLE_CANT_HANDLE_SPKEYS
        ER_UNSUPPORTED_EXTENSION
        ER_UNSUPPORTED_PS
        ER_WARN_ENGINE_TRANSACTION_ROLLBACK
    """,
    _oursqlx_OperationalError="""
        CR_SERVER_GONE_ERROR
        CR_SERVER_LOST
        ER_ABORTING_CONNECTION
        ER_ALTER_FILEGROUP_FAILED
        ER_AUTOINC_READ_FAILED
        ER_BINLOG_PURGE_FATAL_ERR
        ER_BINLOG_ROW_LOGGING_FAILED
        ER_CANNOT_USER
        ER_CANT_CREATE_DB
        ER_CANT_CREATE_FILE
        ER_CANT_CREATE_HANDLER_FILE
        ER_CANT_CREATE_TABLE
        ER_CANT_CREATE_THREAD
        ER_CANT_DELETE_FILE
        ER_CANT_FIND_SYSTEM_REC
        ER_CANT_GET_STAT
        ER_CANT_GET_WD
        ER_CANT_LOCK
        ER_CANT_OPEN_FILE
        ER_CANT_OPEN_LIBRARY
        ER_CANT_READ_DIR
        ER_CANT_SET_WD
        ER_CHECKREAD
        ER_CONFLICT_FN_PARSE_ERROR
        ER_CONNECT_TO_MASTER
        ER_CRASHED_ON_REPAIR
        ER_CRASHED_ON_USAGE
        ER_CREATE_FILEGROUP_FAILED
        ER_DB_DROP_DELETE
        ER_DB_DROP_RMDIR
        ER_DDL_LOG_ERROR
        ER_DISK_FULL
        ER_DROP_FILEGROUP_FAILED
        ER_ERROR_DURING_CHECKPOINT
        ER_ERROR_DURING_COMMIT
        ER_ERROR_DURING_FLUSH_LOGS
        ER_ERROR_DURING_ROLLBACK
        ER_ERROR_ON_CLOSE
        ER_ERROR_ON_READ
        ER_ERROR_ON_RENAME
        ER_ERROR_ON_WRITE
        ER_ERROR_WHEN_EXECUTING_COMMAND
        ER_EVENT_CANNOT_DELETE
        ER_EVENT_CANT_ALTER
        ER_EVENT_COMPILE_ERROR
        ER_EVENT_DROP_FAILED
        ER_EVENT_MODIFY_QUEUE_ERROR
        ER_EVENT_OPEN_TABLE_FAILED
        ER_EVENT_SET_VAR_ERROR
        ER_EVENT_STORE_FAILED
        ER_EXCEPTIONS_WRITE_ERROR
        ER_FAILED_ROUTINE_BREAK_BINLOG
        ER_FILE_NOT_FOUND
        ER_FILE_USED
        ER_FSEEK_FAIL
        ER_GET_ERRMSG
        ER_GET_ERRNO
        ER_GET_TEMPORARY_ERRMSG
        ER_INDEX_REBUILD
        ER_IO_ERR_LOG_INDEX_READ
        ER_LOG_PURGE_UNKNOWN_ERR
        ER_MASTER
        ER_MASTER_FATAL_ERROR_READING_BINLOG
        ER_MASTER_INFO
        ER_MASTER_NET_READ
        ER_MASTER_NET_WRITE
        ER_NET_ERROR_ON_WRITE
        ER_NET_FCNTL_ERROR
        ER_NET_PACKETS_OUT_OF_ORDER
        ER_NET_READ_ERROR
        ER_NET_READ_ERROR_FROM_PIPE
        ER_NET_READ_INTERRUPTED
        ER_NET_UNCOMPRESS_ERROR
        ER_NET_WRITE_INTERRUPTED
        ER_NO_FILE_MAPPING
        ER_PROC_AUTO_GRANT_FAIL
        ER_PROC_AUTO_REVOKE_FAIL
        ER_QUERY_ON_MASTER
        ER_REGEXP_ERROR
        ER_RELAY_LOG_FAIL
        ER_RELAY_LOG_INIT
        ER_SLAVE_THREAD
        ER_SP_CANT_ALTER
        ER_SP_DROP_FAILED
        ER_SP_PROC_TABLE_CORRUPT
        ER_SP_STORE_FAILED
        ER_UNEXPECTED_EOF
        ER_VIEW_CHECKSUM
        ER_WARNING_NOT_COMPLETE_ROLLBACK
        ER_WSAS_FAILED
        ER_XAER_RMERR
    """,
    _oursqlx_PermissionsError="""
        ER_ACCESS_DENIED_ERROR
        ER_BINLOG_CREATE_ROUTINE_NEED_SUPER
        ER_CANT_CREATE_USER_WITH_GRANT
        ER_COLUMNACCESS_DENIED_ERROR
        ER_DBACCESS_DENIED_ERROR
        ER_HOST_NOT_PRIVILEGED
        ER_KILL_DENIED_ERROR
        ER_NO_PERMISSION_TO_CREATE_USER
        ER_PASSWORD_NOT_ALLOWED
        ER_PROCACCESS_DENIED_ERROR
        ER_SPECIFIC_ACCESS_DENIED_ERROR
        ER_TABLEACCESS_DENIED_ERROR
        ER_VIEW_NO_EXPLAIN
        ER_VIEW_OTHER_USER
    """,
)

def main(outfile):
    output = []
    for exc, consts in sorted(data.items()):
        for const in consts.split():
            output.append(case_template % dict(const=const))
        output.append(return_template % dict(exc=exc))
    outfile.write(file_template % dict(
        when=time.asctime(),
        code=''.join(output),
    ))

if __name__ == '__main__':
    main(open(sys.argv[1], 'w'))



# This is definitely my least favorite module in the whole project. I hate you,
# MySQL, I really do. 

cdef class _ResultSet:
    """_ResultSet(...)
    
    You should never ever instantiate a _ResultSet yourself unless you really
    know what you're doing. 
    """
    
    cdef Connection conn
    cdef MYSQL_RES *res
    cdef unsigned int fields
    cdef object conversion_info
    cdef readonly object rowcount, lastrowid, description, column_names
    cdef public object warnings
    cdef int show_table
    
    conversions_dict = _conversions
    
    def __cinit__(self, Connection conn not None, bint show_table=False,
            **kwargs):
        cdef my_ulonglong res
        conn._check_closed()
        self.conn = conn
        self.fields = mysql_field_count(self.conn.conn)
        self.res = mysql_store_result(self.conn.conn)
        if not self.res and self.fields != 0:
            self._raise_error()
        
        res = mysql_affected_rows(self.conn.conn)
        if res == ull_negone:
            self.rowcount = -1
        else:
            self.rowcount = res
        res = mysql_insert_id(self.conn.conn)
        if res == 0:
            self.lastrowid = None
        else:
            self.lastrowid = res
        
        if self.fields:
            self.description = description_from_res(
                conn, self.res, self.fields, self.show_table)
            self.conversion_info = conversion_info_from_res(
                self.res, self.fields)
            self.column_names = [d[0] for d in self.description]
        else:
            self.description = self.column_names = None
        
        if self.conn.warning_count():
            self.warnings = _do_warnings_query(self.conn)
        else:
            self.warnings = []
    
    def __dealloc__(self):
        self.close()
    
    def close(self):
        if self.res:
            mysql_free_result(self.res)
            self.res = NULL
    
    cdef int _raise_error(self) except -1:
        return self.conn._raise_error()
    
    cdef int _check_closed(self) except -1:
        if not self.conn.conn:
            if self.res:
                self.close()
            raise ProgrammingError('connection closed')
        if not self.res:
            raise ProgrammingError('query closed')
    
    property has_result:
        def __get__(self):
            return self.fields != 0
    
    def fetchone(self):
        cdef MYSQL_ROW row
        cdef unsigned int i
        cdef unsigned long *lengths
        self._check_closed()
        row = mysql_fetch_row(self.res)
        if not row:
            return None
        lengths = mysql_fetch_lengths(self.res)
        if not lengths:
            self._raise_error()
        ret = PyTuple_New(self.fields)
        # Cache the connection charset.
        self.conn.charset
        for 0 <= i < self.fields:
            if not row[i]:
                d = None
            else:
                d = PyBytes_FromStringAndSize(row[i], lengths[i])
                field_type, flags, charsetnr = self.conversion_info[i]
                converter = self.conversions_dict.get(field_type)
                if converter:
                    d = converter(d)
                elif self.conn.use_unicode and charsetnr != CS_BINARY:
                    d = d.decode(self.conn._charset)
            Py_INCREF(d)
            PyTuple_SET_ITEM(ret, i, d)
        return ret
    
    def fetchall(self):
        self._check_closed()
        return PySequence_List(self)
    
    def __iter__(self):
        self._check_closed()
        return PyCallIter_New(self.fetchone, None)

cdef class _Query:
    """_Query(...)
    
    _Queries generate _ResultSets. Eventually they will also support generating
    multiple _ResultSets when dealing with queries that produce multiple result
    sets.
    """
    
    cdef Connection conn
    cdef object result, res_a, res_kw
    cdef readonly object rowcount, lastrowid, description, warnings, has_result
    
    def __cinit__(self, Connection conn not None, query, *a, **kw):
        cdef char *qstr
        cdef Py_ssize_t qlen
        conn._check_closed()
        self.conn = conn
        query = bytes_maybe_from_encoding(query, self.conn.charset, False)
        PyBytes_AsStringAndSize(query, &qstr, &qlen)
        if mysql_real_query(self.conn.conn, qstr, qlen):
            self._raise_error()
        self.res_a, self.res_kw = a, kw
        self.warnings = []
        self.new_result()
        # XXX: support multiple result sets.
        self.has_result = self.result.has_result
    
    _result_class = _ResultSet
    def new_result(self):
        res = self._result_class(self.conn, *self.res_a, **self.res_kw)
        self.warnings.extend(res.warnings)
        self.rowcount = res.rowcount
        self.lastrowid = res.lastrowid
        self.description = res.description
        self.result = res
    
    cdef int _raise_error(self) except -1:
        return self.conn._raise_error()
    
    def nextset(self):
        cdef int res
        # XXX: support multiple result sets.
        self.result = None
        return False
        
        if not self.result:
            return False
        self.result.close()
        res = mysql_next_result(self.conn.conn)
        if res > 0:
            self._raise_error()
        elif res == -1:
            self.result = None
            return False
        elif res == 0:
            self.new_result()
            return True
    
    def fetchone(self):
        while self.result is not None:
            if self.result.has_result:
                row = self.result.fetchone()
                if row is not None:
                    return row
            self.nextset()
        return None
    
    def __iter__(self):
        return PyCallIter_New(self.fetchone, None)
    
    def fetchall(self):
        return PySequence_List(self)

# Cython should support multiple inheritance one of these days.
_DictResultSet = type('_DictResultSet', (_DictWhateverMixin, _ResultSet), 
    {'__module__': __name__})

cdef class _DictQuery(_Query):
    _result_class = _DictResultSet



import datetime
import decimal
import codecs
import io

cdef my_ulonglong ull_negone = <my_ulonglong>-1

cdef class _AbstractIterWrapper: 
    def __iter__(self):
        return self

cdef class _BinaryWhateverMixin:
    def __next__(self):
        ret = super(BinaryFileWrapper, self).next()
        return ret

cdef class IterWrapper(_AbstractIterWrapper):
    """IterWrapper(iterobj)
    
    IterWrapper is a wrapper class for arbitrary string-yielding iterable 
    objects so that their contents can be streamed over the wire as a bound 
    parameter for a query. The strings are assumed to be text and not arbitrary
    bytes.
    """
    
    cdef object iterobj, nextfunc
    
    def __cinit__(self, iterobj):
        self.iterobj = iter(iterobj)
        self.nextfunc = self.iterobj.next
    
    def __next__(self):
        return self.nextfunc()

BinaryIterWrapper = type('BinaryIterWrapper', 
    (_BinaryWhateverMixin, IterWrapper), {'__module__': __name__, '__doc__':
    """BinaryIterWrapper(iterobj)
    
    BinaryIterWrapper behaves like IterWrapper, except the strings yielded from
    the iterable are treated as bytes instead of text strings.
    """})

cdef class FileWrapper(_AbstractIterWrapper):
    """FileWrapper(fileobj, chunksize=8192, doclose=False)
    
    FileWrapper is a wrapper class for file-like objects so that their contents
    can be streamed over the wire as a bound parameter for a query. Strings 
    read from the wrapped file are assumed to be text and not arbitrary bytes.
    
    Strings will be read by calling the 'read' method on the file object with
    the 'chunksize' parameter until it returns the empty string. If 'doclose'
    is true, the file will have its 'close' method called once it has been 
    exhausted.
    """
    
    cdef object fileobj, chunksize, readfunc
    cdef int doclose, done
    
    def __cinit__(self, fileobj, chunksize=8192, bint doclose=False):
        self.fileobj = fileobj
        self.readfunc = fileobj.read
        self.chunksize = chunksize
        self.doclose = doclose
        self.done = False
    
    def __next__(self):
        if self.done:
            raise StopIteration
        ret = self.readfunc(self.chunksize)
        if not ret:
            if self.doclose:
                self.fileobj.close()
            self.done = True
            raise StopIteration
        return ret

BinaryFileWrapper = type('BinaryFileWrapper', 
    (_BinaryWhateverMixin, FileWrapper), {'__module__': __name__, '__doc__':
    """BinaryFileWrapper(fileobj, chunksize=8192, doclose=False)
    
    BinaryFileWrapper behaves exactly like FileWrapper, except the strings 
    yielded from the file object are treated as bytes instead of text strings.
    """})

cdef class _Statement:
    """_Statement(...)
    
    A class that represents a prepared statement. You should probably avoid
    touching this, even if it's more stable than _Query/_ResultSet.
    """
    cdef object __weakref__
    cdef MYSQL_STMT *stmt
    cdef MYSQL_BIND *bind
    cdef column_output *result_data
    cdef unsigned int result_fields, fetched
    cdef unsigned long cursor_type, string_limit, _prefetch_rows
    cdef int buffered, data_waiting, show_table, exhausted
    cdef Connection conn
    cdef public object warnings
    
    def __cinit__(self, Connection connection not None, 
            unsigned long string_limit=0, bint show_table=False, **kwargs):
        self.conn = connection
        self.stmt = mysql_stmt_init(self.conn.conn)
        if not self.stmt:
            raise MemoryError('alloc of statement failed')
        self.string_limit = string_limit
        self.show_table = show_table
        self.result_data = NULL
        self.warnings = []
        self.buffered = self.data_waiting = self.exhausted = False
        if _oursqlx_init_stmt_cursor(self.stmt, &self.cursor_type):
            self._raise_error()
    
    def __dealloc__(self):
        self.close()
        PyMem_Free(self.result_data)
        PyMem_Free(self.bind)
    
    def close(self):
        if self.stmt:
            mysql_stmt_close(self.stmt)
            self.stmt = NULL
    
    cdef int _raise_error(self) except -1:
        cdef int err = mysql_stmt_errno(self.stmt)
        # effect a sideeffect by looking up the charset, required for
        # _decode_char_p
        self.conn.charset
        raise _exception_from_errno(err)(
            self.conn._decode_char_p(mysql_stmt_error(self.stmt)), err)
    
    cdef int _check_closed(self) except -1:
        if not self.conn.conn:
            if self.stmt:
                self.close()
            raise ProgrammingError('connection closed')
        if not self.stmt:
            raise ProgrammingError('statement closed')
    
    property string_limit:
        def __get__(self):
            return self.string_limit
        
        def __set__(self, unsigned long value):
            if value < 0:
                raise ValueError('string limit must be >= 0')
            self.string_limit = value
    
    def prepare(self, statement):
        cdef char *buf
        cdef Py_ssize_t bufsize
        self._check_closed()
        if self.data_waiting:
            raise ProgrammingError('binding this query would cause rows to '
                'become unfetchable')
        statement = bytes_maybe_from_encoding(
            statement, self.conn.charset, False)
        PyBytes_AsStringAndSize(statement, &buf, &bufsize)
        if mysql_stmt_prepare(self.stmt, buf, bufsize):
            self._raise_error()
        self._alloc_buffer()
    
    cdef int _alloc_buffer(self) except -1:
        cdef MYSQL_RES *res
        cdef MYSQL_BIND *bind = NULL
        cdef column_output *result_data = NULL
        res = mysql_stmt_result_metadata(self.stmt)
        
        # Okay so. MySQL doesn't require that we keep around the MYSQL_BIND 
        # structure that we pass to mysql_stmt_bind_result, but we do have to
        # keep around the buffers that we've bound to hold the results. If 
        # there's a failure at this point, we're pretty much screwed since the
        # result buffers won't be prepared and we can't restore the old 
        # prepared statement.
        
        # So! Since we have no reason not to, we free any previously bound 
        # buffers.
        PyMem_Free(self.result_data)
        self.result_data = NULL
        # But if we don't need any buffers, we're done anyway.
        if not res:
            return 0
        
        try:
            # Otherwise, we (try to) allocate the result buffers.
            fields = mysql_num_fields(res)
            bind = <MYSQL_BIND *>PyMem_Malloc(sizeof(MYSQL_BIND) * fields)
            if not bind:
                raise MemoryError('alloc of result bind structure failed')
            memset(bind, 0, sizeof(MYSQL_BIND) * fields)
            result_data = <column_output *>PyMem_Malloc(
                sizeof(column_output) * fields)
            if not result_data:
                raise MemoryError('alloc of result buffers failed')
            memset(result_data, 0, sizeof(column_output) * fields)
        except:
            # We only want to free the buffers if there was a failure and we 
            # can't use 'em.
            PyMem_Free(result_data)
            PyMem_Free(bind)
            raise
        else:
            # And we only want to save the buffers if there was no failure.
            self.bind = bind
            self.result_data = result_data
            self.result_fields = fields
        finally:
            # But in any case, we don't need the result.
            mysql_free_result(res)
    
    cdef int _bind_buffer(self) except -1:
        cdef MYSQL_RES *res
        cdef MYSQL_FIELD *field
        cdef MYSQL_BIND *b
        cdef column_output *c
        cdef unsigned int fields, i
        res = mysql_stmt_result_metadata(self.stmt)
        
        try:
            # And now we (try to) set up the result buffers.
            fields = mysql_num_fields(res)
            for 0 <= i < fields:
                b, c = &self.bind[i], &self.result_data[i]
                field = mysql_fetch_field_direct(res, i)
                if field.type == MYSQL_TYPE_GEOMETRY:
                    c.type = b.buffer_type = MYSQL_TYPE_STRING
                else:
                    c.type = b.buffer_type = field.type
                c.flags = field.flags
                c.is_binary = (field.charsetnr == CS_BINARY)
                b.buffer_length = sizeof(column_data)
                b.buffer = &c.buf
                b.length = &c.length
                b.is_null = &c.is_null
                b.error = &c.error
                if field.flags & UNSIGNED_NUM_FLAG == UNSIGNED_NUM_FLAG:
                    b.is_unsigned = True
            if mysql_stmt_bind_result(self.stmt, self.bind):
                self._raise_error()
        finally:
            # And clean up the result data.
            mysql_free_result(res)
    
    property has_result:
        def __get__(self):
            self._check_closed()
            return self.result_data != NULL
    
    property description:
        def __get__(self):
            cdef MYSQL_RES *res
            self._check_closed()
            if not self.result_data:
                return None
            res = mysql_stmt_result_metadata(self.stmt)
            if not res:
                self.result_data = NULL
                return
            try:
                ret = description_from_res(
                    self.conn, res, self.result_fields, self.show_table)
            finally:
                mysql_free_result(res)
            return ret
    
    property column_names:
        def __get__(self):
            desc = self.description
            if desc is None:
                return None
            return [x[0] for x in desc]
    
    def execute(self, *parameters):
        cdef char *buf = NULL
        cdef void *long_buffer
        cdef MYSQL_BIND *bind = NULL, *b
        cdef column_data *column, *c
        cdef Py_ssize_t i, n_params, stmt_params, long_buffer_size
        self._check_closed()
        if self.data_waiting:
            raise ProgrammingError('executing this query would cause rows to '
                'become unfetchable')
        n_params = PyTuple_GET_SIZE(parameters)
        stmt_params = mysql_stmt_param_count(self.stmt)
        if stmt_params != n_params:
            raise ProgrammingError('%d parameters expected, %d given' % (
                stmt_params, n_params))
        long_strings = []
        # So we don't have temporary values cleaned up before they're used.
        temp_strings = []
        try:
            if n_params > 0:
                buf = <char *>PyMem_Malloc(
                    (sizeof(MYSQL_BIND) + sizeof(column_data)) * n_params)
                if not buf:
                    raise MemoryError('alloc of bind parameters failed')
                bind = <MYSQL_BIND *>buf
                memset(bind, 0, sizeof(MYSQL_BIND) * n_params)
                column = <column_data *>(buf + (sizeof(MYSQL_BIND) * n_params))
                memset(column, 0, sizeof(column_data) * n_params)
                # Cache the connection charset.
                self.conn.charset
                for 0 <= i < n_params:
                    o = PyTuple_GET_ITEM(parameters, i)
                    Py_INCREF(o)
                    b, c = &bind[i], &column[i]
                    # This monolith should be refactored into something better
                    # someday. It can't really work the same way as MySQLdb has
                    # a conversion dict, since this involves some C-level union
                    # and buffer munging. 
                    if o is None:
                        b.buffer_type = MYSQL_TYPE_NULL
                        b.buffer = NULL
                    elif PyUnicode_Check(o):
                        b.buffer_type = MYSQL_TYPE_STRING
                        if self.conn.use_unicode:
                            o = o.encode(self.conn._charset)
                            temp_strings.append(o)
                        PyBytes_AsStringAndSize(o, 
                            <char **>&b.buffer, &c.st)
                        c.ul = c.st
                        b.length = &c.ul
                    elif PyBytes_Check(o):
                        b.buffer_type = MYSQL_TYPE_STRING
                        PyBytes_AsStringAndSize(o, 
                            <char **>&b.buffer, &c.st)
                        c.ul = c.st
                        b.length = &c.ul
                    elif PyInt_Check(o):
                        # MYSQL_TYPE_LONG really takes a C int, not a C long.
                        # This is necessary for platforms where sizeof(long) >
                        # sizeof(int).
                        b.buffer_type = MYSQL_TYPE_LONGLONG
                        c.ll = PyInt_AS_LONG(o)
                        b.buffer = <void *>&c.ll
                    elif PyLong_Check(o):
                        b.buffer_type = MYSQL_TYPE_LONGLONG
                        if o >= 0:
                            c.ull = PyLong_AsUnsignedLongLong(o)
                            b.buffer = <void *>&c.ull
                            b.is_unsigned = True
                        else:
                            c.ll = PyLong_AsLongLong(o)
                            b.buffer = <void *>&c.ll
                    elif PyFloat_Check(o):
                        b.buffer_type = MYSQL_TYPE_DOUBLE
                        c.d = PyFloat_AS_DOUBLE(o)
                        b.buffer = <void *>&c.d
                    elif PyDateTime_Check(o):
                        b.buffer_type = MYSQL_TYPE_DATETIME
                        c.t.year = PyDateTime_GET_YEAR(o)
                        c.t.month = PyDateTime_GET_MONTH(o)
                        c.t.day = PyDateTime_GET_DAY(o)
                        c.t.hour = PyDateTime_DATE_GET_HOUR(o)
                        c.t.minute = PyDateTime_DATE_GET_MINUTE(o)
                        c.t.second = PyDateTime_DATE_GET_SECOND(o)
                        c.t.second_part = PyDateTime_DATE_GET_MICROSECOND(o)
                        b.buffer = <void *>&c.t
                    elif PyDate_Check(o):
                        b.buffer_type = MYSQL_TYPE_DATE
                        c.t.year = PyDateTime_GET_YEAR(o)
                        c.t.month = PyDateTime_GET_MONTH(o)
                        c.t.day = PyDateTime_GET_DAY(o)
                        b.buffer = <void *>&c.t
                    elif PyTime_Check(o):
                        b.buffer_type = MYSQL_TYPE_TIME
                        c.t.hour = PyDateTime_TIME_GET_HOUR(o)
                        c.t.minute = PyDateTime_TIME_GET_MINUTE(o)
                        c.t.second = PyDateTime_TIME_GET_SECOND(o)
                        c.t.second_part = PyDateTime_TIME_GET_MICROSECOND(o)
                        b.buffer = <void *>&c.t
                    elif isinstance(o, decimal.Decimal):
                        b.buffer_type = MYSQL_TYPE_NEWDECIMAL
                        o = str(o).encode()
                        temp_strings.append(o)
                        b.buffer = PyBytes_AS_STRING(o)
                        c.ul = PyBytes_GET_SIZE(o)
                        b.length = &c.ul
                    elif isinstance(o, _AbstractIterWrapper):
                        b.buffer_type = MYSQL_TYPE_STRING
                        b.length = &c.ul
                        long_strings.append((i, o))
                    else:
                        raise TypeError('cannot bind %r object' % o.__class__)
            if mysql_stmt_bind_param(self.stmt, bind):
                self._raise_error()
            for i, o in long_strings:
                for s in o:
                    if self.conn.use_unicode:
                        s = s.encode(self.conn._charset)
                    PyBytes_AsStringAndSize(s, 
                        <char **>&long_buffer, &long_buffer_size)
                    if mysql_stmt_send_long_data(self.stmt, i, 
                            <char *>long_buffer, long_buffer_size):
                        self._raise_error()
            if mysql_stmt_execute(self.stmt):
                self._raise_error()
            if _oursqlx_stmt_cursor_prefetch(self.stmt, &self.buffered):
                self._raise_error()
            self.exhausted = False
            if self.result_data:
                self.data_waiting = True
            if self.conn.warning_count():
                self.warnings.extend(_do_warnings_query(self.conn))
        finally:
            PyMem_Free(buf)
    
    def fetchone(self):
        cdef MYSQL_BIND b_tryagain
        cdef column_output *c
        cdef column_data *d
        cdef unsigned int i
        cdef int result
        cdef uint64_t u
        self._check_closed()
        if not self.result_data or self.exhausted:
            return None
        elif self.result_data:
            self._bind_buffer()
        result = mysql_stmt_fetch(self.stmt)
        self.fetched += 1
        if result == 1:
            self._raise_error()
        elif result == MYSQL_NO_DATA:
            if mysql_stmt_free_result(self.stmt):
                self._raise_error()
            self.buffered = self.data_waiting = False
            self.exhausted = True
            return None
        if self.result_fields == 0:
            return ()
        ret = PyTuple_New(self.result_fields)
        # Cache the connection charset.
        self.conn.charset
        for 0 <= i < self.result_fields:
            c = &self.result_data[i]
            d = &c.buf
            if c.is_null:
                val = None
            elif c.type in (MYSQL_TYPE_STRING, MYSQL_TYPE_BLOB, 
                    MYSQL_TYPE_VAR_STRING):
                # Oh string handling.
                val = None
                # If the string wasn't too long to fit in our minimal buffer,
                # then we can just copy it out of the buffer.
                if not c.error:
                    val = PyBytes_FromStringAndSize(&d.c, c.length)
                # If we already have all the results stored clientside, or we
                # don't care about strings being too long, or we *do* care 
                # about strings being too long and it's not too long...
                elif (self.buffered
                        or not self.string_limit 
                        or c.length <= self.string_limit):
                    # First, allocate a string of the necessary length.
                    val = PyBytes_FromStringAndSize(NULL, c.length)
                    # Then, fill up a MYSQL_BIND that will fetch into the 
                    # aforementioned string.
                    memset(&b_tryagain, 0, sizeof(MYSQL_BIND))
                    b_tryagain.buffer_type = MYSQL_TYPE_BLOB
                    b_tryagain.buffer_length = c.length
                    b_tryagain.buffer = PyBytes_AS_STRING(val)
                    # This means that if there is still a truncation error for
                    # whatever reason, something later in the function will
                    # complain.
                    b_tryagain.error = &c.error
                    # Aaand fetch it.
                    if mysql_stmt_fetch_column(self.stmt, &b_tryagain, i, 0):
                        self._raise_error()
                # But, if we care about the length of the string, we need to
                # make sure we have a consistent API.
                if self.string_limit:
                    # At this point, if 'val' still isn't set, it means we 
                    # should stream it from the database.
                    if val is None:
                        val = _ResultStringStream(self, i, c.length)
                        c.error = False
                    # Since _ResultStringStream has a file-like interface, we
                    # make sure that everything passed back has a file-like
                    # interface.
                    else:
                        val = io.BytesIO(val)
                    # And if we're decoding unicode, decode unicode in a 
                    # streaming fashion!
                    if self.conn.use_unicode and not c.is_binary:
                        val = codecs.getreader(self.conn._charset)(val)
                else:
                    # At this point, 'val' should be a string read in from the
                    # database since there's no possibility of streaming. If 
                    # we're handling unicode, decode it.
                    if self.conn.use_unicode and not c.is_binary:
                        val = val.decode(self.conn._charset)
            elif c.type in (MYSQL_TYPE_DECIMAL, MYSQL_TYPE_NEWDECIMAL):
                val = decimal.Decimal(
                PyBytes_FromStringAndSize(d.dec, c.length).decode(self.conn._charset))
            # Sometimes mysql likes to give back invalid data instead of doing
            # anything sensible.
            elif c.type in (MYSQL_TYPE_DATETIME, MYSQL_TYPE_TIMESTAMP, 
                    MYSQL_TYPE_DATE) and d.t.year == 0:
                val = None
            elif c.type in (MYSQL_TYPE_DATETIME, MYSQL_TYPE_TIMESTAMP):
                val = PyDateTime_FromDateAndTime(d.t.year, d.t.month, d.t.day, 
                    d.t.hour, d.t.minute, d.t.second, d.t.second_part)
            elif c.type == MYSQL_TYPE_DATE:
                val = PyDate_FromDate(d.t.year, d.t.month, d.t.day)
            elif c.type == MYSQL_TYPE_TIME:
                val = PyTime_FromTime(
                    d.t.hour, d.t.minute, d.t.second, d.t.second_part)
            elif c.type == MYSQL_TYPE_TINY:
                if c.flags & UNSIGNED_FLAG == UNSIGNED_FLAG:
                    val = PyLong_FromUnsignedLong(d.u8)
                else:
                    val = PyLong_FromLong(d.s8)
            elif c.type in (MYSQL_TYPE_SHORT, MYSQL_TYPE_YEAR):
                if c.flags & UNSIGNED_FLAG == UNSIGNED_FLAG:
                    val = PyLong_FromUnsignedLong(d.u16)
                else:
                    val = PyLong_FromLong(d.s16)
            elif c.type in (MYSQL_TYPE_LONG, MYSQL_TYPE_INT24):
                if c.flags & UNSIGNED_FLAG == UNSIGNED_FLAG:
                    val = PyLong_FromUnsignedLong(d.u32)
                else:
                    val = PyLong_FromLong(d.s32)
            elif c.type == MYSQL_TYPE_LONGLONG:
                if c.flags & UNSIGNED_FLAG == UNSIGNED_FLAG:
                    val = PyLong_FromUnsignedLongLong(d.u64)
                else:
                    val = PyLong_FromLongLong(d.s64)
            elif c.type == MYSQL_TYPE_FLOAT:
                val = d.f
            elif c.type == MYSQL_TYPE_DOUBLE:
                val = d.d
            elif c.type == MYSQL_TYPE_BIT:
                # Bitfields column come variable-length and always in big-
                # endian format. Convert it to an int instead of passing back 
                # a string that people would probably be converting to ints 
                # anyway.
                val = bitval_from_char_p(<unsigned char *>&d.c, c.length)
            else:
                val = PyBytes_FromStringAndSize(&d.c, c.length)
                PyErr_WarnEx(Warning, "unknown column returned as string", 1)
            if c.error:
                # This *shouldn't* ever happen.
                raise InterfaceError('truncated column %d, type %d' % (
                    i, c.type))
            Py_INCREF(val)
            PyTuple_SET_ITEM(ret, i, val)
        return ret
    
    def fetchall(self):
        # buffer_results() calls _check_closed()
        self.buffer_results()
        return PySequence_List(self)
    
    def buffer_results(self):
        self._check_closed()
        if not self.buffered:
            if mysql_stmt_store_result(self.stmt):
                self._raise_error()
            self.buffered = True
    
    def nextset(self):
        return False
    
    def __iter__(self):
        self._check_closed()
        return PyCallIter_New(self.fetchone, None)
    
    property rowcount:
        def __get__(self):
            cdef my_ulonglong res
            self._check_closed()
            res = mysql_stmt_affected_rows(self.stmt)
            if res == ull_negone:
                return -1
            else:
                return res
    
    property lastrowid:
        def __get__(self):
            cdef my_ulonglong res
            self._check_closed()
            res = mysql_stmt_insert_id(self.stmt)
            if res == 0:
                return None
            else:
                return res
    
    property prefetch_rows:
        def __get__(self):
            return self._prefetch_rows
        
        def __set__(self, unsigned long value):
            if value < 1:
                raise ValueError('must prefetch at least 1 row')
            self._prefetch_rows = value
            if _oursqlx_stmt_set_prefetch_rows(self.stmt, 
                    &self._prefetch_rows):
                self._raise_error()

_DictStatement = type('_DictStatement', (_DictWhateverMixin, _Statement), 
    {'__module__': __name__})

cdef class _ResultStringStream:
    """_ResultStringStream(...)
    
    _ResultStringStream represents an incoming stream of binary data from the
    database. You should never ever instantiate this class yourself.
    
    A _ResultStringStream must be used immediately; fetching any additional 
    rows from the database will cause it to go out of sync and become unusable.
    """
    
    cdef MYSQL_BIND bind
    cdef _Statement stmt
    cdef unsigned int index, stmt_fetched
    cdef unsigned long offset, length
    cdef int done
    
    def __cinit__(self, _Statement stmt not None, unsigned int index, 
            unsigned int length):
        self.stmt = stmt
        self.stmt_fetched = stmt.fetched
        self.index = index
        self.length = length
        self.offset = 0
        memset(&self.bind, 0, sizeof(MYSQL_BIND))
        self.bind.buffer_type = MYSQL_TYPE_BLOB
        self.bind.length = &self.length
        self.done = False
    
    def read(self, Py_ssize_t size=-1):
        """read(size=-1) -> str
        
        Read some bytes from the database. Reading -1 bytes will attempt to 
        fetch all remaining bytes.
        """
        cdef Py_ssize_t left
        if self.done:
            return ''
        self.stmt._check_closed()
        if self.stmt.fetched != self.stmt_fetched:
            raise ProgrammingError('row no longer available from database')
        left = self.length - self.offset
        if size == -1 or size > left:
            size = left
        elif size < 0:
            raise ValueError("can't read negative bytes")
        ret = PyBytes_FromStringAndSize(NULL, size)
        self.bind.buffer_length = size
        self.bind.buffer = PyBytes_AS_STRING(ret)
        if mysql_stmt_fetch_column(self.stmt.stmt, 
                &self.bind, self.index, self.offset):
            self.stmt._raise_error()
        self.offset += size
        left = self.length - self.offset
        assert left >= 0, 'size remaining should be >= 0'
        if left == 0:
            self.done = True
        return ret



#!/usr/bin/env python
import time
import sys

file_template = """
/* generated on %(when)s */

PyObject *_oursqlx_generate_errno_dict(void) {
    PyObject *ret = NULL, *val = NULL;
    if (!(ret = PyDict_New()))
        return NULL;
    %(code)s
    return ret;
    
  set_error:
    Py_DECREF(val);
  value_error:
    Py_DECREF(ret);
    return NULL;
}
"""

error_template = """
#ifdef %(error)s
    if (!(val = PyLong_FromLong(%(error)s)))
        goto value_error;
    if (PyDict_SetItemString(ret, "%(error)s", val) == -1)
        goto set_error;
    Py_DECREF(val);
#endif
"""

errors = """
    CR_MIN_ERROR
    CR_MAX_ERROR
    CR_ERROR_FIRST
    CR_UNKNOWN_ERROR
    CR_SOCKET_CREATE_ERROR
    CR_CONNECTION_ERROR
    CR_CONN_HOST_ERROR
    CR_IPSOCK_ERROR
    CR_UNKNOWN_HOST
    CR_SERVER_GONE_ERROR
    CR_VERSION_ERROR
    CR_OUT_OF_MEMORY
    CR_WRONG_HOST_INFO
    CR_LOCALHOST_CONNECTION
    CR_TCP_CONNECTION
    CR_SERVER_HANDSHAKE_ERR
    CR_SERVER_LOST
    CR_COMMANDS_OUT_OF_SYNC
    CR_NAMEDPIPE_CONNECTION
    CR_NAMEDPIPEWAIT_ERROR
    CR_NAMEDPIPEOPEN_ERROR
    CR_NAMEDPIPESETSTATE_ERROR
    CR_CANT_READ_CHARSET
    CR_NET_PACKET_TOO_LARGE
    CR_EMBEDDED_CONNECTION
    CR_PROBE_SLAVE_STATUS
    CR_PROBE_SLAVE_HOSTS
    CR_PROBE_SLAVE_CONNECT
    CR_PROBE_MASTER_CONNECT
    CR_SSL_CONNECTION_ERROR
    CR_MALFORMED_PACKET
    CR_WRONG_LICENSE
    CR_NULL_POINTER
    CR_NO_PREPARE_STMT
    CR_PARAMS_NOT_BOUND
    CR_DATA_TRUNCATED
    CR_NO_PARAMETERS_EXISTS
    CR_INVALID_PARAMETER_NO
    CR_INVALID_BUFFER_USE
    CR_UNSUPPORTED_PARAM_TYPE
    CR_SHARED_MEMORY_CONNECTION
    CR_SHARED_MEMORY_CONNECT_REQUEST_ERROR
    CR_SHARED_MEMORY_CONNECT_ANSWER_ERROR
    CR_SHARED_MEMORY_CONNECT_FILE_MAP_ERROR
    CR_SHARED_MEMORY_CONNECT_MAP_ERROR
    CR_SHARED_MEMORY_FILE_MAP_ERROR
    CR_SHARED_MEMORY_MAP_ERROR
    CR_SHARED_MEMORY_EVENT_ERROR
    CR_SHARED_MEMORY_CONNECT_ABANDONED_ERROR
    CR_SHARED_MEMORY_CONNECT_SET_ERROR
    CR_CONN_UNKNOW_PROTOCOL
    CR_INVALID_CONN_HANDLE
    CR_SECURE_AUTH
    CR_FETCH_CANCELED
    CR_NO_DATA
    CR_NO_STMT_METADATA
    CR_NO_RESULT_SET
    CR_NOT_IMPLEMENTED
    CR_SERVER_LOST_EXTENDED
    CR_STMT_CLOSED
    CR_NEW_STMT_METADATA
    CR_ERROR_LAST
    ER_ERROR_FIRST
    ER_HASHCHK
    ER_NISAMCHK
    ER_NO
    ER_YES
    ER_CANT_CREATE_FILE
    ER_CANT_CREATE_TABLE
    ER_CANT_CREATE_DB
    ER_DB_CREATE_EXISTS
    ER_DB_DROP_EXISTS
    ER_DB_DROP_DELETE
    ER_DB_DROP_RMDIR
    ER_CANT_DELETE_FILE
    ER_CANT_FIND_SYSTEM_REC
    ER_CANT_GET_STAT
    ER_CANT_GET_WD
    ER_CANT_LOCK
    ER_CANT_OPEN_FILE
    ER_FILE_NOT_FOUND
    ER_CANT_READ_DIR
    ER_CANT_SET_WD
    ER_CHECKREAD
    ER_DISK_FULL
    ER_DUP_KEY
    ER_ERROR_ON_CLOSE
    ER_ERROR_ON_READ
    ER_ERROR_ON_RENAME
    ER_ERROR_ON_WRITE
    ER_FILE_USED
    ER_FILSORT_ABORT
    ER_FORM_NOT_FOUND
    ER_GET_ERRNO
    ER_ILLEGAL_HA
    ER_KEY_NOT_FOUND
    ER_NOT_FORM_FILE
    ER_NOT_KEYFILE
    ER_OLD_KEYFILE
    ER_OPEN_AS_READONLY
    ER_OUTOFMEMORY
    ER_OUT_OF_SORTMEMORY
    ER_UNEXPECTED_EOF
    ER_CON_COUNT_ERROR
    ER_OUT_OF_RESOURCES
    ER_BAD_HOST_ERROR
    ER_HANDSHAKE_ERROR
    ER_DBACCESS_DENIED_ERROR
    ER_ACCESS_DENIED_ERROR
    ER_NO_DB_ERROR
    ER_UNKNOWN_COM_ERROR
    ER_BAD_NULL_ERROR
    ER_BAD_DB_ERROR
    ER_TABLE_EXISTS_ERROR
    ER_BAD_TABLE_ERROR
    ER_NON_UNIQ_ERROR
    ER_SERVER_SHUTDOWN
    ER_BAD_FIELD_ERROR
    ER_WRONG_FIELD_WITH_GROUP
    ER_WRONG_GROUP_FIELD
    ER_WRONG_SUM_SELECT
    ER_WRONG_VALUE_COUNT
    ER_TOO_LONG_IDENT
    ER_DUP_FIELDNAME
    ER_DUP_KEYNAME
    ER_DUP_ENTRY
    ER_WRONG_FIELD_SPEC
    ER_PARSE_ERROR
    ER_EMPTY_QUERY
    ER_NONUNIQ_TABLE
    ER_INVALID_DEFAULT
    ER_MULTIPLE_PRI_KEY
    ER_TOO_MANY_KEYS
    ER_TOO_MANY_KEY_PARTS
    ER_TOO_LONG_KEY
    ER_KEY_COLUMN_DOES_NOT_EXITS
    ER_BLOB_USED_AS_KEY
    ER_TOO_BIG_FIELDLENGTH
    ER_WRONG_AUTO_KEY
    ER_READY
    ER_NORMAL_SHUTDOWN
    ER_GOT_SIGNAL
    ER_SHUTDOWN_COMPLETE
    ER_FORCING_CLOSE
    ER_IPSOCK_ERROR
    ER_NO_SUCH_INDEX
    ER_WRONG_FIELD_TERMINATORS
    ER_BLOBS_AND_NO_TERMINATED
    ER_TEXTFILE_NOT_READABLE
    ER_FILE_EXISTS_ERROR
    ER_LOAD_INFO
    ER_ALTER_INFO
    ER_WRONG_SUB_KEY
    ER_CANT_REMOVE_ALL_FIELDS
    ER_CANT_DROP_FIELD_OR_KEY
    ER_INSERT_INFO
    ER_UPDATE_TABLE_USED
    ER_NO_SUCH_THREAD
    ER_KILL_DENIED_ERROR
    ER_NO_TABLES_USED
    ER_TOO_BIG_SET
    ER_NO_UNIQUE_LOGFILE
    ER_TABLE_NOT_LOCKED_FOR_WRITE
    ER_TABLE_NOT_LOCKED
    ER_BLOB_CANT_HAVE_DEFAULT
    ER_WRONG_DB_NAME
    ER_WRONG_TABLE_NAME
    ER_TOO_BIG_SELECT
    ER_UNKNOWN_ERROR
    ER_UNKNOWN_PROCEDURE
    ER_WRONG_PARAMCOUNT_TO_PROCEDURE
    ER_WRONG_PARAMETERS_TO_PROCEDURE
    ER_UNKNOWN_TABLE
    ER_FIELD_SPECIFIED_TWICE
    ER_INVALID_GROUP_FUNC_USE
    ER_UNSUPPORTED_EXTENSION
    ER_TABLE_MUST_HAVE_COLUMNS
    ER_RECORD_FILE_FULL
    ER_UNKNOWN_CHARACTER_SET
    ER_TOO_MANY_TABLES
    ER_TOO_MANY_FIELDS
    ER_TOO_BIG_ROWSIZE
    ER_STACK_OVERRUN
    ER_WRONG_OUTER_JOIN
    ER_NULL_COLUMN_IN_INDEX
    ER_CANT_FIND_UDF
    ER_CANT_INITIALIZE_UDF
    ER_UDF_NO_PATHS
    ER_UDF_EXISTS
    ER_CANT_OPEN_LIBRARY
    ER_CANT_FIND_DL_ENTRY
    ER_FUNCTION_NOT_DEFINED
    ER_HOST_IS_BLOCKED
    ER_HOST_NOT_PRIVILEGED
    ER_PASSWORD_ANONYMOUS_USER
    ER_PASSWORD_NOT_ALLOWED
    ER_PASSWORD_NO_MATCH
    ER_UPDATE_INFO
    ER_CANT_CREATE_THREAD
    ER_WRONG_VALUE_COUNT_ON_ROW
    ER_CANT_REOPEN_TABLE
    ER_INVALID_USE_OF_NULL
    ER_REGEXP_ERROR
    ER_MIX_OF_GROUP_FUNC_AND_FIELDS
    ER_NONEXISTING_GRANT
    ER_TABLEACCESS_DENIED_ERROR
    ER_COLUMNACCESS_DENIED_ERROR
    ER_ILLEGAL_GRANT_FOR_TABLE
    ER_GRANT_WRONG_HOST_OR_USER
    ER_NO_SUCH_TABLE
    ER_NONEXISTING_TABLE_GRANT
    ER_NOT_ALLOWED_COMMAND
    ER_SYNTAX_ERROR
    ER_DELAYED_CANT_CHANGE_LOCK
    ER_TOO_MANY_DELAYED_THREADS
    ER_ABORTING_CONNECTION
    ER_NET_PACKET_TOO_LARGE
    ER_NET_READ_ERROR_FROM_PIPE
    ER_NET_FCNTL_ERROR
    ER_NET_PACKETS_OUT_OF_ORDER
    ER_NET_UNCOMPRESS_ERROR
    ER_NET_READ_ERROR
    ER_NET_READ_INTERRUPTED
    ER_NET_ERROR_ON_WRITE
    ER_NET_WRITE_INTERRUPTED
    ER_TOO_LONG_STRING
    ER_TABLE_CANT_HANDLE_BLOB
    ER_TABLE_CANT_HANDLE_AUTO_INCREMENT
    ER_DELAYED_INSERT_TABLE_LOCKED
    ER_WRONG_COLUMN_NAME
    ER_WRONG_KEY_COLUMN
    ER_WRONG_MRG_TABLE
    ER_DUP_UNIQUE
    ER_BLOB_KEY_WITHOUT_LENGTH
    ER_PRIMARY_CANT_HAVE_NULL
    ER_TOO_MANY_ROWS
    ER_REQUIRES_PRIMARY_KEY
    ER_NO_RAID_COMPILED
    ER_UPDATE_WITHOUT_KEY_IN_SAFE_MODE
    ER_KEY_DOES_NOT_EXITS
    ER_CHECK_NO_SUCH_TABLE
    ER_CHECK_NOT_IMPLEMENTED
    ER_CANT_DO_THIS_DURING_AN_TRANSACTION
    ER_ERROR_DURING_COMMIT
    ER_ERROR_DURING_ROLLBACK
    ER_ERROR_DURING_FLUSH_LOGS
    ER_ERROR_DURING_CHECKPOINT
    ER_NEW_ABORTING_CONNECTION
    ER_DUMP_NOT_IMPLEMENTED
    ER_FLUSH_MASTER_BINLOG_CLOSED
    ER_INDEX_REBUILD
    ER_MASTER
    ER_MASTER_NET_READ
    ER_MASTER_NET_WRITE
    ER_FT_MATCHING_KEY_NOT_FOUND
    ER_LOCK_OR_ACTIVE_TRANSACTION
    ER_UNKNOWN_SYSTEM_VARIABLE
    ER_CRASHED_ON_USAGE
    ER_CRASHED_ON_REPAIR
    ER_WARNING_NOT_COMPLETE_ROLLBACK
    ER_TRANS_CACHE_FULL
    ER_SLAVE_MUST_STOP
    ER_SLAVE_NOT_RUNNING
    ER_BAD_SLAVE
    ER_MASTER_INFO
    ER_SLAVE_THREAD
    ER_TOO_MANY_USER_CONNECTIONS
    ER_SET_CONSTANTS_ONLY
    ER_LOCK_WAIT_TIMEOUT
    ER_LOCK_TABLE_FULL
    ER_READ_ONLY_TRANSACTION
    ER_DROP_DB_WITH_READ_LOCK
    ER_CREATE_DB_WITH_READ_LOCK
    ER_WRONG_ARGUMENTS
    ER_NO_PERMISSION_TO_CREATE_USER
    ER_UNION_TABLES_IN_DIFFERENT_DIR
    ER_LOCK_DEADLOCK
    ER_TABLE_CANT_HANDLE_FT
    ER_CANNOT_ADD_FOREIGN
    ER_NO_REFERENCED_ROW
    ER_ROW_IS_REFERENCED
    ER_CONNECT_TO_MASTER
    ER_QUERY_ON_MASTER
    ER_ERROR_WHEN_EXECUTING_COMMAND
    ER_WRONG_USAGE
    ER_WRONG_NUMBER_OF_COLUMNS_IN_SELECT
    ER_CANT_UPDATE_WITH_READLOCK
    ER_MIXING_NOT_ALLOWED
    ER_DUP_ARGUMENT
    ER_USER_LIMIT_REACHED
    ER_SPECIFIC_ACCESS_DENIED_ERROR
    ER_LOCAL_VARIABLE
    ER_GLOBAL_VARIABLE
    ER_NO_DEFAULT
    ER_WRONG_VALUE_FOR_VAR
    ER_WRONG_TYPE_FOR_VAR
    ER_VAR_CANT_BE_READ
    ER_CANT_USE_OPTION_HERE
    ER_NOT_SUPPORTED_YET
    ER_MASTER_FATAL_ERROR_READING_BINLOG
    ER_SLAVE_IGNORED_TABLE
    ER_INCORRECT_GLOBAL_LOCAL_VAR
    ER_WRONG_FK_DEF
    ER_KEY_REF_DO_NOT_MATCH_TABLE_REF
    ER_OPERAND_COLUMNS
    ER_SUBQUERY_NO_1_ROW
    ER_UNKNOWN_STMT_HANDLER
    ER_CORRUPT_HELP_DB
    ER_CYCLIC_REFERENCE
    ER_AUTO_CONVERT
    ER_ILLEGAL_REFERENCE
    ER_DERIVED_MUST_HAVE_ALIAS
    ER_SELECT_REDUCED
    ER_TABLENAME_NOT_ALLOWED_HERE
    ER_NOT_SUPPORTED_AUTH_MODE
    ER_SPATIAL_CANT_HAVE_NULL
    ER_COLLATION_CHARSET_MISMATCH
    ER_SLAVE_WAS_RUNNING
    ER_SLAVE_WAS_NOT_RUNNING
    ER_TOO_BIG_FOR_UNCOMPRESS
    ER_ZLIB_Z_MEM_ERROR
    ER_ZLIB_Z_BUF_ERROR
    ER_ZLIB_Z_DATA_ERROR
    ER_CUT_VALUE_GROUP_CONCAT
    ER_WARN_TOO_FEW_RECORDS
    ER_WARN_TOO_MANY_RECORDS
    ER_WARN_NULL_TO_NOTNULL
    ER_WARN_DATA_OUT_OF_RANGE
    ER_WARN_USING_OTHER_HANDLER
    ER_CANT_AGGREGATE_2COLLATIONS
    ER_DROP_USER
    ER_REVOKE_GRANTS
    ER_CANT_AGGREGATE_3COLLATIONS
    ER_CANT_AGGREGATE_NCOLLATIONS
    ER_VARIABLE_IS_NOT_STRUCT
    ER_UNKNOWN_COLLATION
    ER_SLAVE_IGNORED_SSL_PARAMS
    ER_SERVER_IS_IN_SECURE_AUTH_MODE
    ER_WARN_FIELD_RESOLVED
    ER_BAD_SLAVE_UNTIL_COND
    ER_MISSING_SKIP_SLAVE
    ER_UNTIL_COND_IGNORED
    ER_WRONG_NAME_FOR_INDEX
    ER_WRONG_NAME_FOR_CATALOG
    ER_WARN_QC_RESIZE
    ER_BAD_FT_COLUMN
    ER_UNKNOWN_KEY_CACHE
    ER_WARN_HOSTNAME_WONT_WORK
    ER_UNKNOWN_STORAGE_ENGINE
    ER_WARN_DEPRECATED_SYNTAX
    ER_NON_UPDATABLE_TABLE
    ER_FEATURE_DISABLED
    ER_OPTION_PREVENTS_STATEMENT
    ER_DUPLICATED_VALUE_IN_TYPE
    ER_TRUNCATED_WRONG_VALUE
    ER_TOO_MUCH_AUTO_TIMESTAMP_COLS
    ER_INVALID_ON_UPDATE
    ER_UNSUPPORTED_PS
    ER_GET_ERRMSG
    ER_GET_TEMPORARY_ERRMSG
    ER_UNKNOWN_TIME_ZONE
    ER_WARN_INVALID_TIMESTAMP
    ER_INVALID_CHARACTER_STRING
    ER_WARN_ALLOWED_PACKET_OVERFLOWED
    ER_CONFLICTING_DECLARATIONS
    ER_SP_NO_RECURSIVE_CREATE
    ER_SP_ALREADY_EXISTS
    ER_SP_DOES_NOT_EXIST
    ER_SP_DROP_FAILED
    ER_SP_STORE_FAILED
    ER_SP_LILABEL_MISMATCH
    ER_SP_LABEL_REDEFINE
    ER_SP_LABEL_MISMATCH
    ER_SP_UNINIT_VAR
    ER_SP_BADSELECT
    ER_SP_BADRETURN
    ER_SP_BADSTATEMENT
    ER_UPDATE_LOG_DEPRECATED_IGNORED
    ER_UPDATE_LOG_DEPRECATED_TRANSLATED
    ER_QUERY_INTERRUPTED
    ER_SP_WRONG_NO_OF_ARGS
    ER_SP_COND_MISMATCH
    ER_SP_NORETURN
    ER_SP_NORETURNEND
    ER_SP_BAD_CURSOR_QUERY
    ER_SP_BAD_CURSOR_SELECT
    ER_SP_CURSOR_MISMATCH
    ER_SP_CURSOR_ALREADY_OPEN
    ER_SP_CURSOR_NOT_OPEN
    ER_SP_UNDECLARED_VAR
    ER_SP_WRONG_NO_OF_FETCH_ARGS
    ER_SP_FETCH_NO_DATA
    ER_SP_DUP_PARAM
    ER_SP_DUP_VAR
    ER_SP_DUP_COND
    ER_SP_DUP_CURS
    ER_SP_CANT_ALTER
    ER_SP_SUBSELECT_NYI
    ER_STMT_NOT_ALLOWED_IN_SF_OR_TRG
    ER_SP_VARCOND_AFTER_CURSHNDLR
    ER_SP_CURSOR_AFTER_HANDLER
    ER_SP_CASE_NOT_FOUND
    ER_FPARSER_TOO_BIG_FILE
    ER_FPARSER_BAD_HEADER
    ER_FPARSER_EOF_IN_COMMENT
    ER_FPARSER_ERROR_IN_PARAMETER
    ER_FPARSER_EOF_IN_UNKNOWN_PARAMETER
    ER_VIEW_NO_EXPLAIN
    ER_FRM_UNKNOWN_TYPE
    ER_WRONG_OBJECT
    ER_NONUPDATEABLE_COLUMN
    ER_VIEW_SELECT_DERIVED
    ER_VIEW_SELECT_CLAUSE
    ER_VIEW_SELECT_VARIABLE
    ER_VIEW_SELECT_TMPTABLE
    ER_VIEW_WRONG_LIST
    ER_WARN_VIEW_MERGE
    ER_WARN_VIEW_WITHOUT_KEY
    ER_VIEW_INVALID
    ER_SP_NO_DROP_SP
    ER_SP_GOTO_IN_HNDLR
    ER_TRG_ALREADY_EXISTS
    ER_TRG_DOES_NOT_EXIST
    ER_TRG_ON_VIEW_OR_TEMP_TABLE
    ER_TRG_CANT_CHANGE_ROW
    ER_TRG_NO_SUCH_ROW_IN_TRG
    ER_NO_DEFAULT_FOR_FIELD
    ER_DIVISION_BY_ZERO
    ER_TRUNCATED_WRONG_VALUE_FOR_FIELD
    ER_ILLEGAL_VALUE_FOR_TYPE
    ER_VIEW_NONUPD_CHECK
    ER_VIEW_CHECK_FAILED
    ER_PROCACCESS_DENIED_ERROR
    ER_RELAY_LOG_FAIL
    ER_PASSWD_LENGTH
    ER_UNKNOWN_TARGET_BINLOG
    ER_IO_ERR_LOG_INDEX_READ
    ER_BINLOG_PURGE_PROHIBITED
    ER_FSEEK_FAIL
    ER_BINLOG_PURGE_FATAL_ERR
    ER_LOG_IN_USE
    ER_LOG_PURGE_UNKNOWN_ERR
    ER_RELAY_LOG_INIT
    ER_NO_BINARY_LOGGING
    ER_RESERVED_SYNTAX
    ER_WSAS_FAILED
    ER_DIFF_GROUPS_PROC
    ER_NO_GROUP_FOR_PROC
    ER_ORDER_WITH_PROC
    ER_LOGGING_PROHIBIT_CHANGING_OF
    ER_NO_FILE_MAPPING
    ER_WRONG_MAGIC
    ER_PS_MANY_PARAM
    ER_KEY_PART_0
    ER_VIEW_CHECKSUM
    ER_VIEW_MULTIUPDATE
    ER_VIEW_NO_INSERT_FIELD_LIST
    ER_VIEW_DELETE_MERGE_VIEW
    ER_CANNOT_USER
    ER_XAER_NOTA
    ER_XAER_INVAL
    ER_XAER_RMFAIL
    ER_XAER_OUTSIDE
    ER_XAER_RMERR
    ER_XA_RBROLLBACK
    ER_NONEXISTING_PROC_GRANT
    ER_PROC_AUTO_GRANT_FAIL
    ER_PROC_AUTO_REVOKE_FAIL
    ER_DATA_TOO_LONG
    ER_SP_BAD_SQLSTATE
    ER_STARTUP
    ER_LOAD_FROM_FIXED_SIZE_ROWS_TO_VAR
    ER_CANT_CREATE_USER_WITH_GRANT
    ER_WRONG_VALUE_FOR_TYPE
    ER_TABLE_DEF_CHANGED
    ER_SP_DUP_HANDLER
    ER_SP_NOT_VAR_ARG
    ER_SP_NO_RETSET
    ER_CANT_CREATE_GEOMETRY_OBJECT
    ER_FAILED_ROUTINE_BREAK_BINLOG
    ER_BINLOG_UNSAFE_ROUTINE
    ER_BINLOG_CREATE_ROUTINE_NEED_SUPER
    ER_EXEC_STMT_WITH_OPEN_CURSOR
    ER_STMT_HAS_NO_OPEN_CURSOR
    ER_COMMIT_NOT_ALLOWED_IN_SF_OR_TRG
    ER_NO_DEFAULT_FOR_VIEW_FIELD
    ER_SP_NO_RECURSION
    ER_TOO_BIG_SCALE
    ER_TOO_BIG_PRECISION
    ER_M_BIGGER_THAN_D
    ER_WRONG_LOCK_OF_SYSTEM_TABLE
    ER_CONNECT_TO_FOREIGN_DATA_SOURCE
    ER_QUERY_ON_FOREIGN_DATA_SOURCE
    ER_FOREIGN_DATA_SOURCE_DOESNT_EXIST
    ER_FOREIGN_DATA_STRING_INVALID_CANT_CREATE
    ER_FOREIGN_DATA_STRING_INVALID
    ER_CANT_CREATE_FEDERATED_TABLE
    ER_TRG_IN_WRONG_SCHEMA
    ER_STACK_OVERRUN_NEED_MORE
    ER_TOO_LONG_BODY
    ER_WARN_CANT_DROP_DEFAULT_KEYCACHE
    ER_TOO_BIG_DISPLAYWIDTH
    ER_XAER_DUPID
    ER_DATETIME_FUNCTION_OVERFLOW
    ER_CANT_UPDATE_USED_TABLE_IN_SF_OR_TRG
    ER_VIEW_PREVENT_UPDATE
    ER_PS_NO_RECURSION
    ER_SP_CANT_SET_AUTOCOMMIT
    ER_MALFORMED_DEFINER
    ER_VIEW_FRM_NO_USER
    ER_VIEW_OTHER_USER
    ER_NO_SUCH_USER
    ER_FORBID_SCHEMA_CHANGE
    ER_ROW_IS_REFERENCED_2
    ER_NO_REFERENCED_ROW_2
    ER_SP_BAD_VAR_SHADOW
    ER_TRG_NO_DEFINER
    ER_OLD_FILE_FORMAT
    ER_SP_RECURSION_LIMIT
    ER_SP_PROC_TABLE_CORRUPT
    ER_SP_WRONG_NAME
    ER_TABLE_NEEDS_UPGRADE
    ER_SP_NO_AGGREGATE
    ER_MAX_PREPARED_STMT_COUNT_REACHED
    ER_VIEW_RECURSIVE
    ER_NON_GROUPING_FIELD_USED
    ER_TABLE_CANT_HANDLE_SPKEYS
    ER_NO_TRIGGERS_ON_SYSTEM_SCHEMA
    ER_REMOVED_SPACES
    ER_AUTOINC_READ_FAILED
    ER_USERNAME
    ER_HOSTNAME
    ER_WRONG_STRING_LENGTH
    ER_NON_INSERTABLE_TABLE
    ER_ADMIN_WRONG_MRG_TABLE
    ER_TOO_HIGH_LEVEL_OF_NESTING_FOR_SELECT
    ER_NAME_BECOMES_EMPTY
    ER_AMBIGUOUS_FIELD_TERM
    ER_FOREIGN_SERVER_EXISTS
    ER_FOREIGN_SERVER_DOESNT_EXIST
    ER_ILLEGAL_HA_CREATE_OPTION
    ER_PARTITION_REQUIRES_VALUES_ERROR
    ER_PARTITION_WRONG_VALUES_ERROR
    ER_PARTITION_MAXVALUE_ERROR
    ER_PARTITION_SUBPARTITION_ERROR
    ER_PARTITION_SUBPART_MIX_ERROR
    ER_PARTITION_WRONG_NO_PART_ERROR
    ER_PARTITION_WRONG_NO_SUBPART_ERROR
    ER_CONST_EXPR_IN_PARTITION_FUNC_ERROR
    ER_NO_CONST_EXPR_IN_RANGE_OR_LIST_ERROR
    ER_FIELD_NOT_FOUND_PART_ERROR
    ER_LIST_OF_FIELDS_ONLY_IN_HASH_ERROR
    ER_INCONSISTENT_PARTITION_INFO_ERROR
    ER_PARTITION_FUNC_NOT_ALLOWED_ERROR
    ER_PARTITIONS_MUST_BE_DEFINED_ERROR
    ER_RANGE_NOT_INCREASING_ERROR
    ER_INCONSISTENT_TYPE_OF_FUNCTIONS_ERROR
    ER_MULTIPLE_DEF_CONST_IN_LIST_PART_ERROR
    ER_PARTITION_ENTRY_ERROR
    ER_MIX_HANDLER_ERROR
    ER_PARTITION_NOT_DEFINED_ERROR
    ER_TOO_MANY_PARTITIONS_ERROR
    ER_SUBPARTITION_ERROR
    ER_CANT_CREATE_HANDLER_FILE
    ER_BLOB_FIELD_IN_PART_FUNC_ERROR
    ER_UNIQUE_KEY_NEED_ALL_FIELDS_IN_PF
    ER_NO_PARTS_ERROR
    ER_PARTITION_MGMT_ON_NONPARTITIONED
    ER_FOREIGN_KEY_ON_PARTITIONED
    ER_DROP_PARTITION_NON_EXISTENT
    ER_DROP_LAST_PARTITION
    ER_COALESCE_ONLY_ON_HASH_PARTITION
    ER_REORG_HASH_ONLY_ON_SAME_NO
    ER_REORG_NO_PARAM_ERROR
    ER_ONLY_ON_RANGE_LIST_PARTITION
    ER_ADD_PARTITION_SUBPART_ERROR
    ER_ADD_PARTITION_NO_NEW_PARTITION
    ER_COALESCE_PARTITION_NO_PARTITION
    ER_REORG_PARTITION_NOT_EXIST
    ER_SAME_NAME_PARTITION
    ER_NO_BINLOG_ERROR
    ER_CONSECUTIVE_REORG_PARTITIONS
    ER_REORG_OUTSIDE_RANGE
    ER_PARTITION_FUNCTION_FAILURE
    ER_PART_STATE_ERROR
    ER_LIMITED_PART_RANGE
    ER_PLUGIN_IS_NOT_LOADED
    ER_WRONG_VALUE
    ER_NO_PARTITION_FOR_GIVEN_VALUE
    ER_FILEGROUP_OPTION_ONLY_ONCE
    ER_CREATE_FILEGROUP_FAILED
    ER_DROP_FILEGROUP_FAILED
    ER_TABLESPACE_AUTO_EXTEND_ERROR
    ER_WRONG_SIZE_NUMBER
    ER_SIZE_OVERFLOW_ERROR
    ER_ALTER_FILEGROUP_FAILED
    ER_BINLOG_ROW_LOGGING_FAILED
    ER_BINLOG_ROW_WRONG_TABLE_DEF
    ER_BINLOG_ROW_RBR_TO_SBR
    ER_EVENT_ALREADY_EXISTS
    ER_EVENT_STORE_FAILED
    ER_EVENT_DOES_NOT_EXIST
    ER_EVENT_CANT_ALTER
    ER_EVENT_DROP_FAILED
    ER_EVENT_INTERVAL_NOT_POSITIVE_OR_TOO_BIG
    ER_EVENT_ENDS_BEFORE_STARTS
    ER_EVENT_EXEC_TIME_IN_THE_PAST
    ER_EVENT_OPEN_TABLE_FAILED
    ER_EVENT_NEITHER_M_EXPR_NOR_M_AT
    ER_COL_COUNT_DOESNT_MATCH_CORRUPTED
    ER_CANNOT_LOAD_FROM_TABLE
    ER_EVENT_CANNOT_DELETE
    ER_EVENT_COMPILE_ERROR
    ER_EVENT_SAME_NAME
    ER_EVENT_DATA_TOO_LONG
    ER_DROP_INDEX_FK
    ER_WARN_DEPRECATED_SYNTAX_WITH_VER
    ER_CANT_WRITE_LOCK_LOG_TABLE
    ER_CANT_LOCK_LOG_TABLE
    ER_FOREIGN_DUPLICATE_KEY
    ER_COL_COUNT_DOESNT_MATCH_PLEASE_UPDATE
    ER_TEMP_TABLE_PREVENTS_SWITCH_OUT_OF_RBR
    ER_STORED_FUNCTION_PREVENTS_SWITCH_BINLOG_FORMAT
    ER_NDB_CANT_SWITCH_BINLOG_FORMAT
    ER_PARTITION_NO_TEMPORARY
    ER_PARTITION_CONST_DOMAIN_ERROR
    ER_PARTITION_FUNCTION_IS_NOT_ALLOWED
    ER_DDL_LOG_ERROR
    ER_NULL_IN_VALUES_LESS_THAN
    ER_WRONG_PARTITION_NAME
    ER_CANT_CHANGE_TX_ISOLATION
    ER_DUP_ENTRY_AUTOINCREMENT_CASE
    ER_EVENT_MODIFY_QUEUE_ERROR
    ER_EVENT_SET_VAR_ERROR
    ER_PARTITION_MERGE_ERROR
    ER_CANT_ACTIVATE_LOG
    ER_RBR_NOT_AVAILABLE
    ER_BASE64_DECODE_ERROR
    ER_EVENT_RECURSION_FORBIDDEN
    ER_EVENTS_DB_ERROR
    ER_ONLY_INTEGERS_ALLOWED
    ER_UNSUPORTED_LOG_ENGINE
    ER_BAD_LOG_STATEMENT
    ER_CANT_RENAME_LOG_TABLE
    ER_WRONG_PARAMCOUNT_TO_NATIVE_FCT
    ER_WRONG_PARAMETERS_TO_NATIVE_FCT
    ER_WRONG_PARAMETERS_TO_STORED_FCT
    ER_NATIVE_FCT_NAME_COLLISION
    ER_DUP_ENTRY_WITH_KEY_NAME
    ER_BINLOG_PURGE_EMFILE
    ER_EVENT_CANNOT_CREATE_IN_THE_PAST
    ER_EVENT_CANNOT_ALTER_IN_THE_PAST
    ER_SLAVE_INCIDENT
    ER_NO_PARTITION_FOR_GIVEN_VALUE_SILENT
    ER_BINLOG_UNSAFE_STATEMENT
    ER_SLAVE_FATAL_ERROR
    ER_SLAVE_RELAY_LOG_READ_FAILURE
    ER_SLAVE_RELAY_LOG_WRITE_FAILURE
    ER_SLAVE_CREATE_EVENT_FAILURE
    ER_SLAVE_MASTER_COM_FAILURE
    ER_BINLOG_LOGGING_IMPOSSIBLE
    ER_VIEW_NO_CREATION_CTX
    ER_VIEW_INVALID_CREATION_CTX
    ER_SR_INVALID_CREATION_CTX
    ER_TRG_CORRUPTED_FILE
    ER_TRG_NO_CREATION_CTX
    ER_TRG_INVALID_CREATION_CTX
    ER_EVENT_INVALID_CREATION_CTX
    ER_TRG_CANT_OPEN_TABLE
    ER_CANT_CREATE_SROUTINE
    ER_SLAVE_AMBIGOUS_EXEC_MODE
    ER_NO_FORMAT_DESCRIPTION_EVENT_BEFORE_BINLOG_STATEMENT
    ER_SLAVE_CORRUPT_EVENT
    ER_LOAD_DATA_INVALID_COLUMN
    ER_LOG_PURGE_NO_FILE
    ER_XA_RBTIMEOUT
    ER_XA_RBDEADLOCK
    ER_NEED_REPREPARE
    ER_DELAYED_NOT_SUPPORTED
    ER_VARIABLE_IS_READONLY
    ER_WARN_ENGINE_TRANSACTION_ROLLBACK
    ER_SLAVE_HEARTBEAT_FAILURE
    ER_SLAVE_HEARTBEAT_VALUE_OUT_OF_RANGE
    ER_NDB_REPLICATION_SCHEMA_ERROR
    ER_CONFLICT_FN_PARSE_ERROR
    ER_EXCEPTIONS_WRITE_ERROR
    ER_TOO_LONG_TABLE_COMMENT
    ER_TOO_LONG_FIELD_COMMENT
    ER_FUNC_INEXISTENT_NAME_COLLISION
    ER_DATABASE_NAME
    ER_TABLE_NAME
    ER_PARTITION_NAME
    ER_SUBPARTITION_NAME
    ER_TEMPORARY_NAME
    ER_RENAMED_NAME
    ER_TOO_MANY_CONCURRENT_TRXS
    ER_ERROR_LAST
"""

def main(outfile):
    output = []
    for error in errors.split():
        output.append(error_template % dict(error=error))
    outfile.write(file_template % dict(
        when=time.asctime(),
        code=''.join(output),
    ))

if __name__ == '__main__':
    main(open(sys.argv[1], 'w'))



# Anything that transmits over the network should be run with nogil. I only 
# wish cython had an easier way of doing this than having all of this stupid
# duplication of functions.

cdef inline MYSQL *mysql_real_connect(
        MYSQL *conn, char *host, char *user, char *passwd, char *db, 
        unsigned int port, char *unix_socket, unsigned long client_flag):
    with nogil:
        return _mysql_real_connect(conn, host, user, passwd, db, port, 
            unix_socket, client_flag)

cdef inline int mysql_ping(MYSQL *conn):
    with nogil:
        return _mysql_ping(conn)

cdef inline my_bool mysql_commit(MYSQL *conn):
    with nogil:
        return _mysql_commit(conn)

cdef inline my_bool mysql_rollback(MYSQL *conn):
    with nogil:
        return _mysql_rollback(conn)

cdef inline int mysql_stmt_prepare(MYSQL_STMT *stmt, 
        char *stmt_str, unsigned long length):
    with nogil:
        return _mysql_stmt_prepare(stmt, stmt_str, length)

cdef inline int mysql_stmt_execute(MYSQL_STMT *stmt):
    with nogil:
        return _mysql_stmt_execute(stmt)

cdef inline int mysql_stmt_fetch(MYSQL_STMT *stmt):
    with nogil:
        return _mysql_stmt_fetch(stmt)

cdef inline my_bool mysql_stmt_send_long_data(MYSQL_STMT *stmt, 
        unsigned int index, char *data, unsigned long length):
    with nogil:
        return _mysql_stmt_send_long_data(stmt, index, data, length)

cdef inline int mysql_stmt_fetch_column(MYSQL_STMT *stmt, 
        MYSQL_BIND *bind, unsigned int column, unsigned long offset):
    with nogil:
        return _mysql_stmt_fetch_column(stmt, bind, column, offset)

cdef inline int mysql_stmt_store_result(MYSQL_STMT *stmt):
    with nogil:
        return _mysql_stmt_store_result(stmt)

cdef inline int mysql_real_query(MYSQL *conn, char *stmt_str, 
        unsigned long length):
    with nogil:
        return _mysql_real_query(conn, stmt_str, length)

cdef inline MYSQL_RES *mysql_use_result(MYSQL *conn):
    with nogil:
        return _mysql_use_result(conn)

cdef inline MYSQL_RES *mysql_store_result(MYSQL *conn):
    with nogil:
        return _mysql_store_result(conn)

cdef inline int mysql_next_result(MYSQL *conn):
    with nogil:
        return _mysql_next_result(conn)

cdef inline MYSQL_ROW mysql_fetch_row(MYSQL_RES *result):
    with nogil:
        return _mysql_fetch_row(result)



cdef extern from *:
    ctypedef char const_char "const char"
    ctypedef char my_bool
    ctypedef struct MYSQL_STMT:
        pass
    ctypedef struct MYSQL:
        pass

cdef extern from "compat.h" nogil:
    enum _oursqlx_exception_type:
        _oursqlx_InterfaceError
        _oursqlx_DataError
        _oursqlx_OperationalError
        _oursqlx_IntegrityError
        _oursqlx_InternalError
        _oursqlx_ProgrammingError
        _oursqlx_NotSupportedError
        _oursqlx_PermissionsError
        _oursqlx_UnknownError
    
    my_bool _oursqlx_init_stmt_cursor(MYSQL_STMT *, unsigned long *)
    my_bool _oursqlx_stmt_set_prefetch_rows(MYSQL_STMT *, unsigned long *)
    int _oursqlx_stmt_cursor_prefetch(MYSQL_STMT *, int *)
    _oursqlx_exception_type _oursqlx_exc_from_errno(int)
    object _oursqlx_escape_string(MYSQL *, object)
    object _oursqlx_generate_errno_dict()
    
    ctypedef signed char int8_t
    ctypedef unsigned char uint8_t
    ctypedef signed int int16_t
    ctypedef unsigned int uint16_t
    ctypedef signed long int int32_t
    ctypedef unsigned long int uint32_t
    ctypedef signed long long int int64_t
    ctypedef unsigned long long int uint64_t

cdef extern from "Python.h":
    object PyCallIter_New(object, object)
    
    object PySequence_List(object)
    object PyList_New(Py_ssize_t)
    void PyList_SET_ITEM(object, Py_ssize_t, object)
    
    object PyTuple_New(Py_ssize_t)
    Py_ssize_t PyTuple_GET_SIZE(object)
    object PyTuple_GET_ITEM(object, Py_ssize_t)
    void PyTuple_SET_ITEM(object, Py_ssize_t, object)
    
    object PyBytes_FromString(void *)
    object PyBytes_FromStringAndSize(void *, Py_ssize_t)
    object PyBytes_FromFormat(char *, ...)
    object PyBytes_FromObject(object)
    int PyBytes_AsStringAndSize(object, char **, Py_ssize_t *) except -1
    int PyBytes_Check(object)
    char *PyBytes_AsString(object) except NULL
    char *PyBytes_AS_STRING(object)
    Py_ssize_t PyBytes_GET_SIZE(object)
    
    bint PyUnicode_Check(object)
    object PyUnicode_FromString(char *)
    object PyUnicode_FromStringAndSize(char *, Py_ssize_t)
    object PyUnicode_AsUTF8String(object)
    object PyUnicode_Decode(char *, Py_ssize_t, char *, char *)
    
    bint PyInt_Check(object)
    long PyInt_AS_LONG(object)
    
    bint PyLong_Check(object)
    long long PyLong_AsLongLong(object) except *
    unsigned long long PyLong_AsUnsignedLongLong(object) except *
    object PyLong_FromLong(long)
    object PyLong_FromUnsignedLong(unsigned long)
    object PyLong_FromLongLong(long long)
    object PyLong_FromUnsignedLongLong(unsigned long long)
    
    bint PyFloat_Check(object)
    double PyFloat_AS_DOUBLE(object)
    
    void PyMem_Free(void *ptr)
    void *PyMem_Malloc(Py_ssize_t)
    
    int PyErr_WarnEx(object, char *, int) except -1
    
    int Py_EQ, Py_NE
    
    # raaaaaauuuuugh
    void Py_INCREF(object)

cdef extern from "datetime.h":
    bint PyDate_Check(object)
    object PyDate_FromDate(int, int, int)
    object PyDate_FromTimestamp(object)
    int PyDateTime_GET_YEAR(object)
    int PyDateTime_GET_MONTH(object)
    int PyDateTime_GET_DAY(object)
    
    bint PyDateTime_Check(object)
    object PyDateTime_FromDateAndTime(int, int, int, int, int, int, int)
    object PyDateTime_FromTimestamp(object)
    int PyDateTime_DATE_GET_HOUR(object)
    int PyDateTime_DATE_GET_MINUTE(object)
    int PyDateTime_DATE_GET_SECOND(object)
    int PyDateTime_DATE_GET_MICROSECOND(object)
    
    bint PyTime_Check(object)
    object PyTime_FromTime(int, int, int, int)
    int PyDateTime_TIME_GET_HOUR(object)
    int PyDateTime_TIME_GET_MINUTE(object)
    int PyDateTime_TIME_GET_SECOND(object)
    int PyDateTime_TIME_GET_MICROSECOND(object)

    int PyDateTime_IMPORT

cdef extern from "string.h":
    int strcmp(char *, char *)
    Py_ssize_t strlen(char *)
    void *memset(void *, int, Py_ssize_t)
    void *memcpy(void *, void *, Py_ssize_t)

cdef extern from "mysql.h" nogil:
    ctypedef unsigned long long my_ulonglong
    
    int MYSQL_NO_DATA
    int MYSQL_DATA_TRUNCATED
    
    ctypedef char **MYSQL_ROW
    ctypedef struct MYSQL_RES: 
        pass
    
    unsigned int NOT_NULL_FLAG
    unsigned int UNSIGNED_FLAG
    unsigned int NUM_FLAG
    
    unsigned long CLIENT_COMPRESS
    unsigned long CLIENT_FOUND_ROWS
    unsigned long CLIENT_MULTI_RESULTS
    unsigned long CLIENT_MULTI_STATEMENTS
    
    # Really, mysql? This magical value isn't defined anywhere?
    DEF CS_BINARY = 63
    
    enum mysql_option:
        MYSQL_INIT_COMMAND
        MYSQL_OPT_CONNECT_TIMEOUT
        MYSQL_OPT_READ_TIMEOUT
        MYSQL_OPT_WRITE_TIMEOUT
        MYSQL_OPT_LOCAL_INFILE
        MYSQL_OPT_NAMED_PIPE
        MYSQL_OPT_PROTOCOL
        MYSQL_READ_DEFAULT_FILE
        MYSQL_READ_DEFAULT_GROUP
        MYSQL_REPORT_DATA_TRUNCATION
        MYSQL_SECURE_AUTH
        MYSQL_SET_CHARSET_DIR
        MYSQL_SET_CHARSET_NAME
        MYSQL_SHARED_MEMORY_BASE_NAME
        MYSQL_OPT_READ_TIMEOUT
        MYSQL_OPT_WRITE_TIMEOUT
        MYSQL_OPT_USE_RESULT
        MYSQL_OPT_USE_REMOTE_CONNECTION
        MYSQL_OPT_USE_EMBEDDED_CONNECTION
        MYSQL_OPT_GUESS_CONNECTION
        MYSQL_SET_CLIENT_IP
        MYSQL_SECURE_AUTH
        MYSQL_REPORT_DATA_TRUNCATION
        MYSQL_OPT_RECONNECT
        MYSQL_OPT_SSL_VERIFY_SERVER_CERT
    
    enum mysql_protocol_type:
        MYSQL_PROTOCOL_DEFAULT
        MYSQL_PROTOCOL_TCP
        MYSQL_PROTOCOL_SOCKET
        MYSQL_PROTOCOL_PIPE
        MYSQL_PROTOCOL_MEMORY
    
    enum enum_field_types:
        MYSQL_TYPE_DECIMAL
        MYSQL_TYPE_TINY
        MYSQL_TYPE_SHORT  
        MYSQL_TYPE_LONG
        MYSQL_TYPE_FLOAT
        MYSQL_TYPE_DOUBLE
        MYSQL_TYPE_NULL
        MYSQL_TYPE_TIMESTAMP
        MYSQL_TYPE_LONGLONG
        MYSQL_TYPE_INT24
        MYSQL_TYPE_DATE
        MYSQL_TYPE_TIME
        MYSQL_TYPE_DATETIME
        MYSQL_TYPE_YEAR
        MYSQL_TYPE_NEWDATE
        MYSQL_TYPE_VARCHAR
        MYSQL_TYPE_BIT
        MYSQL_TYPE_NEWDECIMAL
        MYSQL_TYPE_ENUM
        MYSQL_TYPE_SET
        MYSQL_TYPE_TINY_BLOB
        MYSQL_TYPE_MEDIUM_BLOB
        MYSQL_TYPE_LONG_BLOB
        MYSQL_TYPE_BLOB
        MYSQL_TYPE_VAR_STRING
        MYSQL_TYPE_STRING
        MYSQL_TYPE_GEOMETRY
    
    ctypedef struct MYSQL_BIND:
        enum_field_types buffer_type
        void *buffer
        unsigned long buffer_length
        unsigned long *length
        my_bool *is_null
        my_bool is_unsigned
        my_bool *error
    
    ctypedef struct MYSQL_TIME:
        unsigned int year, month, day, hour, minute, second
        unsigned long second_part
    
    ctypedef struct MYSQL_FIELD:
        char *name, *table
        unsigned int flags, length, max_length, decimals, charsetnr
        enum_field_types type
    
    MYSQL *mysql_init(MYSQL *)
    int mysql_options(MYSQL *, mysql_option, char *)
    my_bool mysql_ssl_set(MYSQL *, char *key, char *cert, char *ca, 
        char *capath, char *cipher)
    MYSQL *_mysql_real_connect "mysql_real_connect" (
        MYSQL *, char *host, char *user, char *passwd, char *db, 
        unsigned int port, char *unix_socket, unsigned long client_flag)
    unsigned int mysql_errno(MYSQL *)
    char *mysql_error(MYSQL *)
    int _mysql_ping "mysql_ping" (MYSQL *)
    my_bool _mysql_commit "mysql_commit" (MYSQL *)
    my_bool _mysql_rollback "mysql_rollback" (MYSQL *)
    char *mysql_character_set_name(MYSQL *)
    int mysql_set_character_set(MYSQL *, char *csname)
    unsigned int mysql_warning_count(MYSQL *)
    void mysql_close(MYSQL *)
    
    MYSQL_STMT *mysql_stmt_init(MYSQL *)
    unsigned int mysql_stmt_errno(MYSQL_STMT *)
    char *mysql_stmt_error(MYSQL_STMT *)
    int _mysql_stmt_prepare "mysql_stmt_prepare" (MYSQL_STMT *, 
        char *stmt_str, unsigned long length)
    my_bool mysql_stmt_bind_param(MYSQL_STMT *, MYSQL_BIND *bind)
    my_bool mysql_stmt_bind_result(MYSQL_STMT *, MYSQL_BIND *bind)
    unsigned long mysql_stmt_param_count(MYSQL_STMT *)
    int _mysql_stmt_execute "mysql_stmt_execute" (MYSQL_STMT *)
    int _mysql_stmt_fetch "mysql_stmt_fetch" (MYSQL_STMT *)
    my_ulonglong mysql_stmt_affected_rows(MYSQL_STMT *)
    my_ulonglong mysql_stmt_insert_id(MYSQL_STMT *)
    my_bool _mysql_stmt_send_long_data "mysql_stmt_send_long_data" (MYSQL_STMT *, 
        unsigned int parameter_number, char *data, unsigned long length)
    int _mysql_stmt_fetch_column "mysql_stmt_fetch_column" (MYSQL_STMT *, 
        MYSQL_BIND *bind, unsigned int column, unsigned long offset)
    int _mysql_stmt_store_result "mysql_stmt_store_result" (MYSQL_STMT *)
    my_bool mysql_stmt_free_result(MYSQL_STMT *)
    void mysql_stmt_close(MYSQL_STMT *)
    
    MYSQL_RES *mysql_stmt_result_metadata(MYSQL_STMT *)
    unsigned int mysql_num_fields(MYSQL_RES *)
    MYSQL_FIELD *mysql_fetch_field_direct(MYSQL_RES *, unsigned int fieldnr)
    void mysql_free_result(MYSQL_RES *)
    
    int _mysql_real_query "mysql_real_query" (MYSQL *, char *stmt_str, 
        unsigned long length)
    unsigned int mysql_field_count(MYSQL *)
    MYSQL_RES *_mysql_use_result "mysql_use_result" (MYSQL *)
    MYSQL_RES *_mysql_store_result "mysql_store_result" (MYSQL *)
    int _mysql_next_result "mysql_next_result" (MYSQL *)
    my_ulonglong mysql_insert_id(MYSQL *)
    my_ulonglong mysql_affected_rows(MYSQL *)
    MYSQL_ROW _mysql_fetch_row "mysql_fetch_row" (MYSQL_RES *)
    unsigned long *mysql_fetch_lengths(MYSQL_RES *)
    
    char *mysql_get_client_info()
    char *mysql_get_server_info(MYSQL *)
    char *mysql_get_ssl_cipher(MYSQL *)
    
    int mysql_library_init(int argc, char **argv, char **groups)

cdef extern from "mysqld_error.h":
    int ER_UNSUPPORTED_PS



cdef object bytes_maybe_from_encoding(object s, object encoding, 
        bint allow_none=True):
    if s is None and allow_none:
        return None
    elif PyUnicode_Check(s):
        s = s.encode(encoding)
    return PyBytes_FromObject(s)

cdef char *bytes_or_null(object s):
    if s is None:
        return NULL
    return PyBytes_AS_STRING(s)

cdef unsigned long bitval_from_char_p(unsigned char *s, Py_ssize_t length):
    cdef Py_ssize_t i
    cdef unsigned long ret = 0
    for 0 <= i < length:
        ret = (ret << 8) | s[i]
    return ret

cdef object conversion_info_from_res(MYSQL_RES *res, int fields):
    cdef int i
    cdef MYSQL_FIELD *field
    ret = PyList_New(fields)
    for 0 <= i < fields:
        field = mysql_fetch_field_direct(res, i)
        data = (
            field.type,
            field.flags,
            field.charsetnr)
        Py_INCREF(data)
        PyList_SET_ITEM(ret, i, data)
    return ret

cdef object description_from_res(Connection conn, MYSQL_RES *res, int fields, 
        bint show_table):
    cdef int i
    cdef MYSQL_FIELD *field
    ret = PyList_New(fields)
    # Cache the charset.
    conn.charset
    for 0 <= i < fields:
        field = mysql_fetch_field_direct(res, i)
        name = conn._decode_char_p(field.name)
        if show_table and field.table[0] != 0:
            name = u'%s.%s' % (conn._decode_char_p(field.table), name)
        data = (
            name,
            field.type,
            field.max_length,
            field.length,
            field.length,
            field.decimals,
            field.flags & NOT_NULL_FLAG != NOT_NULL_FLAG)
        Py_INCREF(data)
        PyList_SET_ITEM(ret, i, data)
    return ret

cdef class _DictWhateverMixin:
    def fetchone(self):
        row = super(_DictWhateverMixin, self).fetchone()
        if row is None:
            return None
        ret = {}
        for name, value in zip(self.column_names, row):
            if name in ret:
                if ret[name] != value:
                    raise ProgrammingError('column "%s" appears more than once'
                        ' in output' % name)
            else:
                ret[name] = value
        return ret

_exception_enum_map = {
    _oursqlx_InterfaceError: InterfaceError,
    _oursqlx_DataError: DataError,
    _oursqlx_OperationalError: OperationalError,
    _oursqlx_IntegrityError: IntegrityError,
    _oursqlx_InternalError: InternalError,
    _oursqlx_ProgrammingError: ProgrammingError,
    _oursqlx_NotSupportedError: NotSupportedError,
    _oursqlx_PermissionsError: PermissionsError,
    _oursqlx_UnknownError: UnknownError,
}
def _exception_from_errno(int err):
    return _exception_enum_map[_oursqlx_exc_from_errno(err)]

def _do_warnings_query(Connection conn not None):
    cdef int should_raise = False
    q = _Query(conn, 'SHOW WARNINGS')
    ret = []
    for kind, err, message in q:
        if kind == 'Note':
            kind = Note
        elif kind == 'Warning':
            kind = Warning
            should_raise = True
        elif kind == 'Error':
            kind = _exception_from_errno(err)
            should_raise = True
        else:
            raise InterfaceError('unknown kind of warning: %r' % kind)
        ret.append((kind, (message, err)))
    if should_raise and conn.raise_on_warnings:
        raise CollatedWarningsError('query caused warnings', extra=ret)
    return ret



_protocol_enum_map = {
    None: MYSQL_PROTOCOL_DEFAULT,
    'tcp': MYSQL_PROTOCOL_TCP,
    'socket': MYSQL_PROTOCOL_SOCKET,
    'pipe': MYSQL_PROTOCOL_PIPE,
    'memory': MYSQL_PROTOCOL_MEMORY,
}

cdef class Connection:
    """Connection([host, user, passwd,] db=None, port=0, unix_socket=None, 
        connect_timeout=None, read_timeout=None, write_timeout=None, 
        protocol=None, ssl=None, read_default_file=None, 
        read_default_group=None, init_command=None, charset_dir=None, 
        shared_memory_base_name=None, local_infile=False, 
        require_secure_auth=False, report_truncation=True, compress=False, 
        found_rows=True, use_unicode=True, charset='utf8', autoping=False, 
        default_cursor=None, raise_on_warnings=True, autoreconnect=False)
    
    Connect to the MySQL database. The first three parameters may be passed by 
    position, but the rest may only be passed by keyword. All options except 
    'use_unicode', 'charset', 'autoping', 'default_cursor', and 
    'raise_on_warnings' correspond with parameters or flags to the 
    mysql_real_connect function, or options to the mysql_options function. 
    
    'protocol' can be one of None, 'tcp', 'socket', 'pipe', or 'memory'. A
    'protocol' of None will use the default protocol, and values otherwise 
    correspond with the potential values of the 'protocol' option in the option 
    file.
    
    'ssl' can either be None or a dict where each key corresponds to a 
    parameter of the mysql_ssl_set function, e.g. {'capath': '/etc/ssl/ca', 
    'cipher': 'DHE-RSA-AES256-SHA'}. Unspecified keys are the same as passing
    NULL to that parameter.
    
    If 'use_unicode' is true, oursql will attempt to return unicode wherever 
    feasible. If 'charset' is not None, oursql will set the connection charset 
    to the value provided. 'charset' defaults to using utf-8. If 'autoping' is 
    true, oursql will run the 'ping' method before executing queries or 
    fetching results to attempt to prevent connection timeout errors. 
    'default_cursor', if provided, will be used as the default cursor class 
    when calling the 'cursor' method. If 'raise_on_warnings' is true, oursql
    will raise a CollatedWarningsError exception if there were any warnings
    produced by running a query.
    
    Connections and their cursors can also be used as context managers.
    
        with some_connection.cursor() as cursor:
            do_something_with(cursor)
    
    Can also be written as:
    
        with some_connection as cursor:
            do_something_with(cursor)
    
    And both are equivalent to:
    
        cursor = some_connection.cursor()
        try:
            do_something_with(cursor)
        except:
            some_connection.rollback()
            raise
        else:
            some_connection.commit()
        finally:
            cursor.close()
    
    """
    
    cdef object __weakref__
    cdef MYSQL *conn
    cdef object _charset, _charset_bytes, _context_cursors
    cdef readonly int use_unicode, autoping, raise_on_warnings
    cdef public object default_cursor
    
    def __cinit__(self, host=None, user=None, passwd=None, *, db=None, 
            unsigned int port=0, unix_socket=None, connect_timeout=None, 
            read_timeout=None, write_timeout=None, protocol=None, ssl=None, 
            read_default_file=None, read_default_group=None, init_command=None, 
            charset_dir=None, shared_memory_base_name=None, 
            bint local_infile=False, bint require_secure_auth=False, 
            bint compress=False, bint report_truncation=True, 
            bint found_rows=True, bint use_unicode=True, charset='utf8', 
            bint autoping=False, default_cursor=None, 
            bint raise_on_warnings=True, bint multi_results=True, 
            bint multi_statements=True, bint autoreconnect=False):
        cdef unsigned long flags = 0
        cdef unsigned int uint_tmp
        cdef my_bool bool_tmp
        if compress:
            flags |= CLIENT_COMPRESS
        if found_rows:
            flags |= CLIENT_FOUND_ROWS
        # I'll come back to this later. Dealing with multiple result sets is
        # such a pain.
        #if multi_results:
        #    flags |= CLIENT_MULTI_RESULTS
        #if multi_statements:
        #    flags |= CLIENT_MULTI_STATEMENTS
        self.conn = mysql_init(NULL)
        if not self.conn:
            raise MemoryError('alloc of conn object failed')
        if protocol not in _protocol_enum_map:
            raise ProgrammingError('unknown protocol %r' % (protocol,))
        host = bytes_maybe_from_encoding(host, charset)
        user = bytes_maybe_from_encoding(user, charset)
        passwd = bytes_maybe_from_encoding(passwd, charset)
        db = bytes_maybe_from_encoding(db, charset)
        unix_socket = bytes_maybe_from_encoding(unix_socket, charset)
        uint_tmp = _protocol_enum_map[protocol]
        mysql_options(self.conn, MYSQL_OPT_PROTOCOL, <char *>&uint_tmp)
        bool_tmp = report_truncation
        mysql_options(self.conn, 
            MYSQL_REPORT_DATA_TRUNCATION, <char *>&bool_tmp)
        bool_tmp = require_secure_auth
        mysql_options(self.conn, MYSQL_SECURE_AUTH, <char *>&bool_tmp)
        if connect_timeout is not None:
            uint_tmp = connect_timeout
            mysql_options(self.conn, 
                MYSQL_OPT_CONNECT_TIMEOUT, <char *>&uint_tmp)
        if read_timeout is not None:
            uint_tmp = read_timeout
            mysql_options(self.conn, 
                MYSQL_OPT_READ_TIMEOUT, <char *>&uint_tmp)
        if write_timeout is not None:
            uint_tmp = write_timeout
            mysql_options(self.conn, 
                MYSQL_OPT_WRITE_TIMEOUT, <char *>&uint_tmp)
        if read_default_file:
            read_default_file = bytes_maybe_from_encoding(
                read_default_file, charset)
            mysql_options(self.conn, 
                MYSQL_READ_DEFAULT_FILE, bytes_or_null(read_default_file))
        if read_default_group:
            read_default_group = bytes_maybe_from_encoding(
                read_default_group, charset)
            mysql_options(self.conn, 
                MYSQL_READ_DEFAULT_GROUP, bytes_or_null(read_default_group))
        if init_command:
            init_command = bytes_maybe_from_encoding(init_command, charset)
            mysql_options(self.conn, 
                MYSQL_INIT_COMMAND, bytes_or_null(init_command))
        if charset_dir:
            charset_dir = bytes_maybe_from_encoding(charset_dir, charset)
            mysql_options(self.conn, 
                MYSQL_INIT_COMMAND, bytes_or_null(charset_dir))
        if shared_memory_base_name:
            shared_memory_base_name = bytes_maybe_from_encoding(
                shared_memory_base_name, charset)
            mysql_options(self.conn, 
                MYSQL_INIT_COMMAND, bytes_or_null(shared_memory_base_name))
        if local_infile:
            mysql_options(self.conn, MYSQL_OPT_LOCAL_INFILE, NULL)
        if ssl is not None:
            self._setup_ssl_options(charset, **ssl)
        if charset is not None:
            charset = PyUnicode_AsUTF8String(charset)
            mysql_options(self.conn, 
                MYSQL_SET_CHARSET_NAME, PyBytes_AS_STRING(charset))
        if not mysql_real_connect(self.conn, bytes_or_null(host), 
                bytes_or_null(user), bytes_or_null(passwd), bytes_or_null(db), 
                port, bytes_or_null(unix_socket), flags):
            self._raise_error()
        if autoreconnect:
            bool_tmp = True
            mysql_options(self.conn, MYSQL_OPT_RECONNECT, <char *>&bool_tmp)
        self.charset
        self.use_unicode = use_unicode
        self.autoping = autoping
        if default_cursor is None:
            default_cursor = Cursor
        self.default_cursor = default_cursor
        self.raise_on_warnings = raise_on_warnings
        self._context_cursors = []
    
    def _setup_ssl_options(self, charset, *, key=None, cert=None, ca=None, 
            capath=None, cipher=None):
        key = bytes_maybe_from_encoding(key, charset)
        cert = bytes_maybe_from_encoding(cert, charset)
        ca = bytes_maybe_from_encoding(ca, charset)
        capath = bytes_maybe_from_encoding(capath, charset)
        cipher = bytes_maybe_from_encoding(cipher, charset)
        mysql_ssl_set(self.conn, bytes_or_null(key), bytes_or_null(cert), 
            bytes_or_null(ca), bytes_or_null(capath), bytes_or_null(cipher))
    
    def close(self):
        """close()
        
        Close the connection. The connection object and all associated cursors
        will become unusable.
        """
        if self.conn:
            self.rollback()
            mysql_close(self.conn)
            self.conn = NULL
    
    def __dealloc__(self):
        if self.conn:
            mysql_close(self.conn)
            self.conn = NULL
    
    cdef int _raise_error(self) except -1:
        # _raise_error can only be called internally, so we shouldn't need to
        # check for if the connection is closed.
        cdef int err = mysql_errno(self.conn)
        self.charset
        raise _exception_from_errno(err)(
            self._decode_char_p(mysql_error(self.conn)), err)
    
    cdef int _check_closed(self) except -1:
        if not self.conn:
            raise ProgrammingError('connection closed')
    
    def ping(self):
        """ping()
        
        Make sure that the database connection is still open. If not, silently
        reconnect.
        """
        self._check_closed()
        if mysql_ping(self.conn):
            self._raise_error()
    
    def commit(self):
        """commit()
        
        Commit the current transaction.
        """
        self._check_closed()
        if mysql_commit(self.conn):
            self._raise_error()
    
    def rollback(self):
        """rollback()
        
        Roll back the current transaction.
        """
        self._check_closed()
        if mysql_rollback(self.conn):
            self._raise_error()
    
    property charset:
        """charset -> str
        
        Get or set the connection's current encoding. If use_unicode is 
        enabled, this is the encoding that will be used to decode incoming
        strings.
        """
        def __get__(self):
            self._charset_bytes = PyBytes_FromString(
                mysql_character_set_name(self.conn))
            self._charset = PyUnicode_FromStringAndSize(
                PyBytes_AS_STRING(self._charset_bytes),
                PyBytes_GET_SIZE(self._charset_bytes))
            return self._charset
        def __set__(self, value):
            self._check_closed()
            svalue = PyUnicode_AsUTF8String(value)
            if mysql_set_character_set(self.conn, PyBytes_AS_STRING(svalue)):
                self._raise_error()
            self._charset_bytes = svalue
            self._charset = value
    
    cdef object _decode_char_p(self, const_char *value):
        cdef Py_ssize_t length = strlen(value)
        return PyUnicode_Decode(value, length, 
            PyBytes_AS_STRING(self._charset_bytes), "strict")
    
    def warning_count(self):
        """warning_count() -> int
        
        Return the number of warnings caused by the most-previously-executed
        query or statement.
        """
        return mysql_warning_count(self.conn)
    
    def cursor(self, cursor_class=None, **kwargs):
        """cursor([cursor_class,] **cursor_options) -> Cursor.
        
        Create a new cursor associated with this connection of the specified 
        type. If no cursor class is specified, the 'default_cursor' attribute
        is used as the cursor class. The cursor_options will be forwarded to 
        the cursor's initializer; see the documentation for the cursor classes 
        for details.
        """
        self._check_closed()
        if cursor_class is None:
            cursor_class = self.default_cursor
        return cursor_class(self, **kwargs)
    
    property server_info:
        def __get__(self):
            cdef const_char *info
            self._check_closed()
            self.charset
            return self._decode_char_p(mysql_get_server_info(self.conn))
    
    property ssl_cipher:
        def __get__(self):
            cdef const_char *res
            self._check_closed()
            res = mysql_get_ssl_cipher(self.conn)
            if not res:
                return None
            self.charset
            return self._decode_char_p(res)
    
    # Unless someone is monkeying around with calling __enter__ and __exit__
    # manually, this approach is the simplest. Yaaay mostly-reentrant code.
    def __enter__(self):
        curs = self.cursor()
        self._context_cursors.append(curs)
        return curs.__enter__()
    
    def __exit__(self, exc, value, tb):
        curs = self._context_cursors.pop()
        curs.__exit__(exc, value, tb)
    
    # FML.
    def _escape_string(self, value):
        """_escape_string(value) -> str.
        
        A wrapper around mysql_real_escape_string. Takes a bytestring in and 
        produces a bytestring out. Please do not every use this unless mysql
        refuses to parameterize a query; oursql has good support for 
        parameterization which also works consistently with unicode.
        """
        return _oursqlx_escape_string(self.conn, value)
    
    def _escape_unicode_string(self, value):
        """_escape_unicode_stirng(value) -> unicode.
        
        _escape_string that takes unicode in and produces unicode out, using
        the connection's charset. As with _escape_string, please do not use 
        this function.
        """
        charset = self.charset
        return self._escape_string(value.encode(charset)).decode(charset)



# cython: language_level=3
"""oursql python bindings for MySQL.

This module is a new, better set of MySQL bindings for python, supporting novel
features like parameterization!
"""

__author__ = 'Aaron Gallagher <habnabit@gmail.com>'
__version__ = '0.9.4'

include "oursql.pxi"
include "nogil.pyx"

PyDateTime_IMPORT

cdef unsigned int UNSIGNED_NUM_FLAG = UNSIGNED_FLAG | NUM_FLAG

cdef union column_data:
    MYSQL_TIME t
    long long ll
    unsigned long long ull
    unsigned long ul
    Py_ssize_t st
    long l
    float f
    double d
    char c
    char dec[65]
    
    int8_t s8
    int16_t s16
    int32_t s32
    int64_t s64
    
    uint8_t u8
    uint16_t u16
    uint32_t u32
    uint64_t u64

cdef union bitfield:
    char c[8]
    uint64_t u64

cdef struct column_output:
    enum_field_types type
    unsigned int flags
    bint is_binary
    unsigned long length
    my_bool is_null
    my_bool error
    column_data buf

import builtins

class Warning(builtins.Warning): pass
class Note(Warning): pass

class Error(builtins.Exception):
    def __init__(self, message, errno=None, extra=None):
        super(Error, self).__init__(errno, message, extra)
        self.errno = errno
        self.extra = extra

class InterfaceError(Error): pass
class DatabaseError(Error): pass
class DataError(DatabaseError): pass
class OperationalError(DatabaseError): pass
class IntegrityError(DatabaseError): pass
class InternalError(DatabaseError): pass
class ProgrammingError(DatabaseError): pass
class CollatedWarningsError(ProgrammingError): pass
class NotSupportedError(DatabaseError): pass
class PermissionsError(DatabaseError): pass

# This exception type is specifically for _exception_from_errno. I think this 
# is a good compromise: if the errno you get isn't valid, return something to 
# indicate that while also being functional.
class UnknownError(Error): pass

cdef class Connection

include "util.pyx"
include "conversions.pyx"
include "connection.pyx"
include "statement.pyx"
include "query.pyx"
include "cursor.pyx"

# The global import lock is held while modules are being imported for the first
# time, so it's threadsafe to call this.
if mysql_library_init(0, NULL, NULL):
    raise InterfaceError('could not initialize mysql')

paramstyle = 'qmark'
threadsafety = 1
apilevel = '2.0'
connect = Connection

cdef class _DBAPITypeObject:
    cdef readonly object values
    
    def __init__(self, *values):
        self.values = frozenset(values)
    
    def __richcmp__(self, other, int op):
        if op == Py_EQ:
            return other in self.values
        elif op == Py_NE:
            return other not in self.values
        else:
            raise TypeError('unorderable types: %r and %r' % (
                type(self), type(other)))
    
    def __repr__(self):
        return '<_DBAPITypeObject representing %r at %#x>' % (
            self.values, id(self))

STRING = _DBAPITypeObject(
    MYSQL_TYPE_ENUM, MYSQL_TYPE_STRING, MYSQL_TYPE_VAR_STRING)
BINARY = _DBAPITypeObject(
    MYSQL_TYPE_BLOB, MYSQL_TYPE_LONG_BLOB, 
    MYSQL_TYPE_MEDIUM_BLOB, MYSQL_TYPE_TINY_BLOB)
NUMBER = _DBAPITypeObject(
    MYSQL_TYPE_DECIMAL, MYSQL_TYPE_DOUBLE, MYSQL_TYPE_FLOAT, MYSQL_TYPE_INT24, 
    MYSQL_TYPE_LONG, MYSQL_TYPE_LONGLONG, MYSQL_TYPE_TINY, MYSQL_TYPE_YEAR)
DATE = _DBAPITypeObject(
    MYSQL_TYPE_DATE, MYSQL_TYPE_NEWDATE)
TIME = _DBAPITypeObject(
    MYSQL_TYPE_TIME)
DATETIME = TIMESTAMP = _DBAPITypeObject(
    MYSQL_TYPE_TIMESTAMP, MYSQL_TYPE_DATETIME)
ROWID = _DBAPITypeObject()

client_info = PyUnicode_FromString(mysql_get_client_info())
errnos = _oursqlx_generate_errno_dict()



import collections, itertools

cdef class Cursor:
    """Cursor(...)
    
    Cursors should never be instantiated directly. They may be passed other 
    options via the Connection.cursor method:
    
      - try_plain_query=True: If this is true, oursql will first attempt to run
        queries as prepared statements, and then fall back to the plain query 
        API if MySQL refuses to parameterize the query. Otherwise, oursql will 
        only run plain queries if explicitly asked.
    
      - show_table=False: If this is true, the name of the table will be 
        prepended to the name of the column in the 'description' attribute of
        the cursor.
    
      - string_limit=0: Set a limit on the size of the buffer oursql will 
        allocate when fetching strings over the wire. If the string_limit is 
        nonzero, file-like objects will always be returned instead of strings. 
        If the size of the string value is greater than the string_limit, the 
        string will be streamed as necessary instead of fetched all at once.
        Note: plain queries buffer incoming result differently, and are 
        therefore unaffected by the string_limit.
    
    Cursor instances yield tuples to represent rows.
    """
    
    cdef readonly object _statements, _stmt_kwargs
    cdef Connection conn
    cdef int closed, try_plain_query
    cdef readonly object rowcount, lastrowid, description, messages
    cdef public unsigned long arraysize
    
    _statement_class = _Statement
    _query_class = _Query
    
    def __cinit__(self, Connection conn not None, bint try_plain_query=True, 
            **kwargs):
        self.conn = conn
        self._stmt_kwargs = kwargs
        self._statements = collections.deque()
        self.messages = []
        self.closed = False
        self.rowcount = -1
        self.arraysize = 1
        self.try_plain_query = try_plain_query
    
    cdef int _check_closed(self) except -1:
        if not self.conn.conn:
            raise ProgrammingError('connection closed')
        if self.closed:
            raise ProgrammingError('cursor closed')
    
    def close(self):
        """close()
        
        Close the cursor and discard the rest of the results. The cursor will
        immediately become unusable.
        """
        self.closed = True
        self._statements.clear()
    
    def _new_statement(self):
        stmt = self._statement_class(self.conn, **self._stmt_kwargs)
        stmt.prefetch_rows = self.arraysize
        return stmt
    
    def _new_query(self, query):
        return self._query_class(self.conn, query, **self._stmt_kwargs)
    
    def callproc(self, procname, params=()):
        """callproc(procname, params=())
        
        Supposedly, run a stored procedure. MySQL's C API has no direct call
        for this; you need to execute stored procedures with the CALL statement
        instead. Calling this method always raises an exception.
        """
        # This is slightly irritating, but not too horribly bad. The horribly 
        # bad part is getting the results from a procedure.
        raise NotSupportedError('use the CALL statement with the execute '
            'method to call stored procedures')
    
    def _do_autoping(self):
        if self.conn.autoping:
            self.conn.ping()
    
    def execute(self, query, params=(), plain_query=False):
        """execute(query, params=(), plain_query=False)
        
        Execute a query. If plain_query is true, run a plain, unparameterized
        query. MySQL cannot parameterize all kinds of statements, which makes
        it sometimes necessary to execute a plain query. If plain_query is true
        and params is true, an exception will be raised as plain queries cannot
        be parameterized.
        """
        self._check_closed()
        self._do_autoping()
        # This, on the other hand, is just plain stupid.
        if plain_query:
            if params:
                # I *really* don't want to write an escaping API. Anything 
                # worth parameterizing supports the prepared statement API
                # anyway, so this shouldn't matter too much.
                raise ProgrammingError("can't call execute with both "
                    "parameters and a plain query")
            stmt = self._new_query(query)
        else:
            stmt = self._new_statement()
            try:
                stmt.prepare(query)
            except Error, e:
                # So, if MySQL refuses to parameterize the query, we optionally
                # try it again through the stupid query API. This is optional
                # because you can lose state data if this is always done 
                # implicitly.
                if self.try_plain_query and e.errno == ER_UNSUPPORTED_PS:
                    return self.execute(query, params, plain_query=True)
                else:
                    raise
            else:
                stmt.execute(*params)
        self._update_fields(stmt)
    
    def executemany(self, query, parambatch):
        """executemany(query, parambatch)
        
        Execute the same query with different sets of parameters. This is more
        efficient than calling execute many times with the same query.
        """
        self._check_closed()
        self._do_autoping()
        # No plain queries because they can't be parameterized anyway.
        stmt = self._new_statement()
        stmt.prepare(query)
        rowcount = 0
        for params in parambatch:
            stmt.execute(*params)
            # There's probably a better way to do this, but whatever. As far as
            # I know we should get all values -1 or all values >= 0.
            rowcount += stmt.rowcount
        # Whatevs.
        if rowcount < 0:
            rowcount = -1
        self._update_fields(stmt, rowcount)
    
    cdef int _update_fields(self, stmt, rowcount=None) except -1:
        self.messages.extend(stmt.warnings)
        if rowcount is None:
            self.rowcount = stmt.rowcount
        else:
            self.rowcount = rowcount
        self.lastrowid = stmt.lastrowid
        self.description = stmt.description
        if stmt.has_result:
            self._statements.append(stmt)
    
    cdef int _check_statements(self) except -1:
        self._check_closed()
        if not self._statements:
            raise ProgrammingError('no results available')
    
    def nextset(self):
        """nextset() -> True or None.
        
        Move to the next result set. If there are any result sets left, this
        returns True. Otherwise, None.
        """
        self._check_statements()
        if self._statements[0].nextset():
            return True
        self._statements.popleft()
        # DB-API, you so wacky.
        return True if self._statements else None
    
    def fetchone(self):
        """fetchone() -> row.
        
        Fetch one row from the current result set. If the result set was
        exhausted, this returns None. If there were no result sets, a 
        ProgrammingError is raised.
        """
        self._check_statements()
        self._do_autoping()
        ret = self._statements[0].fetchone()
        if ret is None:
            self.nextset()
        return ret
    
    def fetchmany(self, size=None):
        """fetchmany(size=self.arraysize) -> list of rows.
        
        Fetch up to 'size' rows. If there aren't that many rows available in 
        the database, this returns as many as were available. If there were no 
        result sets, a ProgrammingError is raised.
        """
        cdef unsigned int i
        self._check_statements()
        self._do_autoping()
        if size is None:
            size = self.arraysize
        ret = []
        resultset = self._statements[0]
        # This is a little bit wacky for DB-API compliance. If there's fewer 
        # rows available than requested, return everything available without
        # switching to the next result set. If there wasn't anything in the
        # result set at all, *then* switch to the next result set and return an
        # empty sequence.
        for 0 <= i < size:
            row = resultset.fetchone()
            if row is None:
                if i == 0:
                    self.nextset()
                return ret
            ret.append(row)
        return ret
    
    def fetchall(self):
        """fetchall() -> list of rows.
        
        Fetch all available rows from the database. If there were no result 
        sets, a ProgrammingError is raised.
        """
        self._check_statements()
        self._do_autoping()
        ret = []
        while self._statements:
            ret.extend(self._statements.popleft().fetchall())
        return ret
    
    def __iter__(self):
        self._check_closed()
        return PyCallIter_New(self.fetchone, None)
    
    property connection:
        def __get__(self):
            return self.conn
    
    def setinputsizes(self, sizes):
        """setinputsizes(sizes)
        
        Do nothing; this operation is not supported by oursql.
        """
        # While I could implement these functions to do something, I don't see
        # it as being very useful for MySQL's C API.
        pass
    
    def setoutputsize(self, size, column=None):
        """setoutputsize(size, column=None)
        
        Do nothing; this operation is not supported by oursql.
        """
        pass
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc, value, tb):
        if exc:
            self.conn.rollback()
        else:
            self.conn.commit()
        self.close()

cdef class DictCursor(Cursor):
    """DictCursor(...)
    
    Cursors should never be instantiated directly. They may be passed other 
    options via the Connection.cursor method. See the documentation for the
    Cursor class for details.
    
    DictCursor instances represent rows as dicts mapping column names or 
    aliases to values. These key names are taken from the 'description' 
    attribute of the cursor, and are therefore affected by the 'show_table'
    cursor option. A ProgrammingError is raised when fetching a row if there 
    were duplicate keys with inequal values.
    """
    
    _statement_class = _DictStatement
    _query_class = _DictQuery



import decimal, datetime, time

def _bitval_from_string(s):
    cdef Py_ssize_t length
    cdef unsigned char *ss
    PyBytes_AsStringAndSize(s, <char **>&ss, &length)
    return bitval_from_char_p(ss, length)

def _datetime_from_string(s):
    tt = time.strptime(s, '%Y-%m-%d %H:%M:%S')
    return datetime.datetime(*tt[:6])

def _date_from_string(s):
    tt = time.strptime(s, '%Y-%m-%d')
    return datetime.date(*tt[:3])

def _time_from_string(s):
    tt = time.strptime(s, '%H:%M:%S')
    return datetime.time(*tt[3:6])

_conversions = {
    MYSQL_TYPE_TINY: int,
    MYSQL_TYPE_SHORT: int,
    MYSQL_TYPE_LONG: long,
    MYSQL_TYPE_FLOAT: float,
    MYSQL_TYPE_DOUBLE: float,
    MYSQL_TYPE_DECIMAL: float,
    MYSQL_TYPE_NEWDECIMAL: float,
    MYSQL_TYPE_LONGLONG: long,
    MYSQL_TYPE_INT24: int,
    MYSQL_TYPE_YEAR: int,
    MYSQL_TYPE_DECIMAL: decimal.Decimal,
    MYSQL_TYPE_NEWDECIMAL: decimal.Decimal,
    MYSQL_TYPE_BIT: _bitval_from_string,
    MYSQL_TYPE_TIMESTAMP: _datetime_from_string,
    MYSQL_TYPE_DATETIME: _datetime_from_string,
    MYSQL_TYPE_DATE: _date_from_string,
    MYSQL_TYPE_TIME: _time_from_string,
    
    # None means to try to decode if we're using unicode and the field isn't 
    # in 'binary' encoding.
    MYSQL_TYPE_BLOB: None,
    MYSQL_TYPE_STRING: None,
    MYSQL_TYPE_VAR_STRING: None,
    MYSQL_TYPE_VARCHAR: None,
    # For consistency with the statement API.
    MYSQL_TYPE_SET: None,
}

Timestamp = datetime.datetime
Date = datetime.date
Time = datetime.time
Binary = bytes

def TimestampFromTicks(ticks):
    """TimestampFromTicks(ticks) -> Timestamp.
    
    Convert a time in seconds since the epoch into a Timestamp.
    """
    return Timestamp(*time.localtime(ticks)[:6])

def DateFromTicks(ticks):
    """DateFromTicks(ticks) -> Date.
    
    Convert a time in seconds since the epoch into a Date.
    """
    return Date(*time.localtime(ticks)[:3])

def TimeFromTicks(ticks):
    """TimeFromTicks(ticks) -> Time.
    
    Convert a time in seconds since the epoch into a Time.
    """
    return Time(*time.localtime(ticks)[3:6])



from distutils.cmd import Command
from distutils.core import setup
from distutils.util import split_quoted
from distutils.extension import Extension
from distutils.msvccompiler import MSVCCompiler
from distutils.command.build_ext import build_ext
import subprocess
import zipfile
import os

try:
    import winreg
except ImportError:
    winreg = None

class MysqlExtension(Extension):
    def __init__(self, *a, **kw):
        self.use_mysql_flags = True
        self._extra_compile_args = []
        self._extra_link_args = []
        self._mysql_compile_args = []
        self._mysql_link_args = []
        self.get_mysql_compile_args = self.get_mysql_link_args = None
        Extension.__init__(self, *a, **kw)
    
    def _get_extra_compile_args(self):
        if not self._mysql_compile_args and self.get_mysql_compile_args:
            self._mysql_compile_args = self.get_mysql_compile_args()
        return self._extra_compile_args + self._mysql_compile_args
    
    def _set_extra_compile_args(self, args):
        self._extra_compile_args = args
    
    extra_compile_args = property(
        _get_extra_compile_args, _set_extra_compile_args)
    
    def _get_extra_link_args(self):
        if not self._mysql_link_args and self.get_mysql_link_args:
            self._mysql_link_args = self.get_mysql_link_args()
        return self._extra_link_args + self._mysql_link_args
    
    def _set_extra_link_args(self, args):
        self._extra_link_args = args
    
    extra_link_args = property(_get_extra_link_args, _set_extra_link_args)

oursql_ext = MysqlExtension("oursql", ["oursqlx/compat.c"])

try:
    from Cython.Distutils import build_ext
except ImportError:
    print("cython not found, using previously-cython'd .c file.")
    oursql_ext.sources.insert(0, 'oursqlx/oursql.c')
else:
    oursql_ext.sources.insert(0, 'oursqlx/oursql.pyx')

oursql_commands = {}

class oursql_build_ext(build_ext):
    user_options = build_ext.user_options + [
        ('mysql-config=', None, 
            '(*nix only) path to the mysql-config executable'),
        ('use-libmysqld', None,
            '(*nix only) link against libmysqld instead of libmysqlclient'),
        ('static', None,
            '(Windows only) statically link against mysqlclient.lib'),
        ('mysql-registry-key=', None,
            '(Windows only) the registry key to query for the mysql root '
            'directory'),
        ('mysql-root=', None,
            '(Windows only) the path to the mysql installation; can be given '
            'instead of --mysql-registry-key'),
    ]
    
    boolean_options = build_ext.boolean_options + [
        'use-libmysqld', 'static'
    ]
    
    def initialize_options(self):
        build_ext.initialize_options(self)
        self.mysql_config = os.environ.get('MYSQL_CONFIG', 'mysql_config')
        self.use_libmysqld = self.static = 0
        self.mysql_registry_key = r'SOFTWARE\MySQL AB\MySQL Server 5.0'
        self.mysql_root = None
    
    def get_mysql_config(self, option):
        args = [self.mysql_config, '--%s' % option]
        print(' '.join(args))
        try:
            proc = subprocess.Popen(args, stdout=subprocess.PIPE)
        except:
            print('failed to execute', args[0])
            raise
        stdout, _ = proc.communicate()
        return split_quoted(stdout.strip().decode())
    
    def setup_posixish(self, ext):
        ext.get_mysql_compile_args = (
            lambda: self.get_mysql_config('cflags'))
        if self.use_libmysqld:
            ext.get_mysql_link_args = (
                lambda: self.get_mysql_config('libmysqld-libs'))
        else:
            ext.get_mysql_link_args = (
                lambda: self.get_mysql_config('libs'))
    
    def setup_windowsish(self, ext):
        if self.mysql_root:
            mysql_root = self.mysql_root
        else:
            mysql_key = winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE, self.mysql_registry_key)
            mysql_root, _ = winreg.QueryValueEx(mysql_key, 'Location')
        
        if self.static:
            client = "mysqlclient"
        else:
            client = "libmysql"

        ext.library_dirs.append(
            os.path.join(mysql_root, 'lib', 'opt'))
        ext.libraries.extend(['advapi32', 'wsock32', client])
        ext.include_dirs.append(
            os.path.join(mysql_root, 'include'))
        if isinstance(self.compiler, MSVCCompiler):
            if self.static:
                ext.get_mysql_compile_args = lambda: ['/MT']
            else:
                ext.get_mysql_compile_args = lambda: ['/MD']

    
    def build_extension(self, ext):
        if getattr(ext, 'use_mysql_flags', False):
            try:
                # distribute/setuptools flip out if I try to open(os.devnull),
                # so, I work around their stupidity by using a pipe instead.
                proc = subprocess.Popen([self.mysql_config], 
                    stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            except OSError:
                if winreg:
                    self.setup_windowsish(ext)
                else:
                    print('warning: no usable mysql_config and no winreg '
                        'module to try; hopefully you have usable '
                        'CFLAGS/LDFLAGS set.')
            else:
                proc.communicate()
                self.setup_posixish(ext)
        build_ext.build_extension(self, ext)
oursql_commands['build_ext'] = oursql_build_ext

try:
    from distutils.command.bdist_msi import bdist_msi
except ImportError:
    pass
else:
    class oursql_bdist_msi(bdist_msi):
        def run(self):
            bdist_msi.run(self)
            if hasattr(self.distribution, 'dist_files'):
                command, pyversion, filename = (
                    self.distribution.dist_files.pop())
                filename = self.get_installer_filename(filename)
                self.distribution.dist_files.append(
                    (command, pyversion, filename))
    oursql_commands['bdist_msi'] = oursql_bdist_msi

class ZipWithLicense(Command):
    description = 'zip binary installer with the license'
    user_options = []
    boolean_options = []
    
    def initialize_options(self):
        pass
    
    def finalize_options(self):
        pass
    
    def run(self):
        license = open(os.path.join(
            os.path.dirname(os.path.abspath(__file__)), 'COPYING')).read()
        license = '\r\n'.join(license.splitlines())
        fullname = self.distribution.get_fullname()
        
        dist_files = getattr(self.distribution, 'dist_files', None)
        if dist_files is None:
            dist_files = []
            for command, cmd_obj in self.distribution.command_obj.iteritems():
                if command.startswith('bdist_'):
                    dist_files.append(
                        (None, None, cmd_obj.get_installer_filename(fullname)))
        
        new_dist_files = []
        for command, pyversion, filename in dist_files:
            out_filename = os.path.splitext(filename)[0] + '.zip'
            out_file = zipfile.ZipFile(out_filename, 'w')
            out_file.write(filename, '%s/%s' % (
                fullname, os.path.basename(filename)))
            out_file.writestr('%s/LICENSE.txt' % fullname, license)
            out_file.close()
            new_dist_files.append((command, pyversion, out_filename))
        self.distribution.dist_files = new_dist_files
oursql_commands['zipwithlicense'] = ZipWithLicense

setup(
    name='oursql3',
    version='0.9.4',
    author='Aaron Gallagher',
    author_email='habnabit@gmail.com',
    url='http://launchpad.net/oursql',
    description='MySQL bindings for python.',
    long_description="""
        oursql is a set of MySQL bindings for python 2.4+ with a focus on 
        wrapping the `MYSQL_STMT API`__ to provide real parameterization and 
        real server-side cursors. MySQL 4.1.2 or better is required.
        
        __ http://dev.mysql.com/doc/refman/5.0/en/c-api-prepared-statements.html
        
        There's extensive documentation available online at 
        http://packages.python.org/oursql/.
    """,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: C',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 3',
        'Topic :: Database :: Database Engines/Servers',
    ],
    
    ext_modules=[oursql_ext],
    cmdclass=oursql_commands,
)
